/*
 "The contents of this file are subject to the Mozilla Public License
 
 Version 1.1 (the "License"); you may not use this file except in
 
 compliance with the License. You may obtain a copy of the License at
 
 http://www.mozilla.org/MPL/
 
 
 Software distributed under the License is distributed on an "AS IS"
 
 basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 
 License for the specific language governing rights and limitations
 
 under the License.
 
 
 The Original Code is OpenMCDF 1.5.4.
 
 
 The Initial Developer of the Original Code is Federico Blaseotto.
 
 Portions created by Federico Blaseotto are Copyright (C)2010 ~ 2013 Federico Blaseotto.
 
 All Rights Reserved.
 
 
 Contributor(s): Autodesk, Inc.."
*/

//
//  FHeader.h
//  LIBCDF
//
//  Created by Weihua Luo on 4/3/2013.
//
#pragma once
#include "IFs/IFObject.h"
#include "FileStream.h"
#include "RWStream.h"

class CFHeader : public RefObject
{
public:
	CFHeader(void);
	CFHeader(unsigned short version);
	~CFHeader(void);

	virtual bool Equals(const IFObject* obj) const {return false;};
	virtual void Delete(){delete this;};

private:
	void Init();
    //0 8 Compound document file identifier: D0H CFH 11H E0H A1H B1H 1AH E1H
	BYTE8 _headerSignature;
    // { 0xD0, 0xCF, 0x11, 0xE0, 0xA1, 0xB1, 0x1A, 0xE1 };

    //8 16 Unique identifier (UID) of this file (not of interest in the following, may be all 0)
    BYTE16 _clsid;

    //24 2 Revision number of the file format (most used is 003EH)
    unsigned short _minorVersion;// = 0x003E;
    //26 2 Version number of the file format (most used is 0003H)
    unsigned short _majorVersion;// = 0x0003;
    //28 2 Byte order identifier (4.2): FEH FFH = Little-Endian FFH FEH = Big-Endian
    unsigned short _byteOrder;// = 0xFFFE;
    //30 2 Size of a sector in the compound document file (3.1) in power-of-two (ssz), real sector
    //size is sec_size = 2ssz bytes (minimum value is 7 which means 128 bytes, most used 
    //value is 9 which means 512 bytes)
    unsigned short _sectorShift;// = 9;
    //32 2 Size of a short-sector in the short-stream container stream (6.1) in power-of-two (sssz),
    //real short-sector size is short_sec_size = 2sssz bytes (maximum value is sector size
    //ssz, see above, most used value is 6 which means 64 bytes)
    unsigned short _miniSectorShift;// = 6;
    //34 6 Not used
    unsigned char _unUsed[6];// = new byte[6];
    //40 4 Total number of sectors used Directory (5.2)
    int _directorySectorsNumber;
    //44 4 Total number of sectors used for the sector allocation table (5.2)
    int _fatSectorsNumber;
    //48 4 SecID of first sector of the directory stream (7)
    int _firstDirectorySectorID;// = Sector.ENDOFCHAIN;
    //52 4 Not used
    unsigned int _unUsed2;
    //56 4 Minimum size of a standard stream (in bytes, minimum allowed and most used size is 4096
    //bytes), streams with an actual size smaller than (and not equal to) this value are stored as
    //short-streams (6)
    unsigned int _minSizeStandardStream;// = 4096;
    //60 4 SecID of first sector of the short-sector allocation table (6.2), or C2 (End Of Chain
    //SecID, 3.1) if not extant
    int _firstMiniFATSectorID;// = unchecked((int)0xFFFFFFFE);
    //64 4 Total number of sectors used for the short-sector allocation table (6.2)
    unsigned int _miniFatSectorsNumber;
    //68 4 SecID of first sector of the master sector allocation table (5.1), or C2 (End Of Chain
    //SecID, 3.1) if no additional sectors used
    int _firstDIFATSectorID;// = Sector.ENDOFCHAIN;
    //72 4 Total number of sectors used for the master sector allocation table (5.1)
    unsigned int _difatSectorsNumber;
    //76 436 First part of the master sector allocation table (5.1) containing 109 SecIDs
    int _difat[109];
public:
	const BYTE8& HeaderSignature() const
    {
        return _headerSignature;
    }

	PROPERTY(BYTE16, CLSID, _clsid);

	PROPERTY_GET(unsigned short, MinorVersion,_minorVersion);

	PROPERTY_GET(unsigned short, MajorVersion, _majorVersion);

	PROPERTY_GET(unsigned short, ByteOrder, _byteOrder);

    PROPERTY_GET(unsigned short, SectorShift, _sectorShift);
          
    PROPERTY_GET(unsigned short, MiniSectorShift, _miniSectorShift);

	unsigned char* UnUsed() const {return (unsigned char*)_unUsed;};

    PROPERTY(int, DirectorySectorsNumber, _directorySectorsNumber);
    PROPERTY(int, FatSectorsNumber, _fatSectorsNumber);
	void IncreaseFatSectorsNumber()
	{
		_fatSectorsNumber++;
	}

    PROPERTY(int, FirstDirectorySectorID,_firstDirectorySectorID);

	int UnUsed2() const {return _unUsed2;};

    PROPERTY(unsigned int, MinSizeStandardStream, _minSizeStandardStream);

    /// <summary>
    /// This integer field contains the starting sector number for the mini FAT
    /// </summary>
    PROPERTY(int, FirstMiniFATSectorID, _firstMiniFATSectorID);

    PROPERTY(unsigned int, MiniFatSectorsNumber, _miniFatSectorsNumber);

    PROPERTY(int, FirstDIFATSectorID, _firstDIFATSectorID);

    PROPERTY(unsigned int, DifatSectorsNumber, _difatSectorsNumber);

	int* DIFAT() const {return (int*)_difat;};

    void Write(CFileStream* stream)
    {
        CRWStream rw(stream);

        rw.Write(_headerSignature.byte, 8);
        rw.Write(_clsid.byte, 16);
        rw.Write(_minorVersion);
        rw.Write(_majorVersion);
        rw.Write(_byteOrder);
        rw.Write(_sectorShift);
        rw.Write(_miniSectorShift);
        rw.Write(_unUsed, 6);
        rw.Write(_directorySectorsNumber);
        rw.Write(_fatSectorsNumber);
        rw.Write(_firstDirectorySectorID);
        rw.Write(_unUsed2);
        rw.Write(_minSizeStandardStream);
        rw.Write(_firstMiniFATSectorID);
        rw.Write(_miniFatSectorsNumber);
        rw.Write(_firstDIFATSectorID);
        rw.Write(_difatSectorsNumber);

        for (int i =0; i<109; i++)
        {
            rw.Write(_difat[i]);
        }

        if (_majorVersion == 4)
        {
            unsigned char zeroHead[3584];
            rw.Write(zeroHead, 3584);
        }

        rw.Close();
    }

    void Read(CFileStream* stream)
    {
        CRWStream rw(stream);

        rw.ReadBytes(_headerSignature.byte, 8);
        CheckSignature();
        rw.ReadBytes(_clsid.byte, 16);
        _minorVersion = rw.ReadUInt16();
        _majorVersion = rw.ReadUInt16();
        CheckVersion();
        _byteOrder = rw.ReadUInt16();
        _sectorShift = rw.ReadUInt16();
        _miniSectorShift = rw.ReadUInt16();
        rw.ReadBytes(_unUsed, 6);
        _directorySectorsNumber = rw.ReadInt32();
        _fatSectorsNumber = rw.ReadInt32();
        _firstDirectorySectorID = rw.ReadInt32();
        _unUsed2 = rw.ReadUInt32();
        _minSizeStandardStream = rw.ReadUInt32();
        _firstMiniFATSectorID = rw.ReadInt32();
        _miniFatSectorsNumber = rw.ReadUInt32();
        _firstDIFATSectorID = rw.ReadInt32();
        _difatSectorsNumber = rw.ReadUInt32();

        for (int i = 0; i < 109; i++)
        {
            _difat[i] = rw.ReadInt32();
        }

        rw.Close();
    }

	void InitSignature()
	{
		unsigned char cs[8] = { 0xD0, 0xCF, 0x11, 0xE0, 0xA1, 0xB1, 0x1A, 0xE1 };
		memcpy(OLE_CFS_SIGNATURE.byte, cs, 8);
	}

private:
    void CheckVersion()
    {
        if (_majorVersion != 3 && _majorVersion != 4)
            throw CFFileFormatException("Unsupported Binary File Format version: This lib only supports Compound Files with major version equal to 3 or 4 ");
    }

    /// <summary>
    /// Structured Storage signature
    /// </summary>
	BYTE8 OLE_CFS_SIGNATURE;

    void CheckSignature()
    {
        for (int i = 0; i < 8; i++)
        {
			if (_headerSignature.byte[i] != OLE_CFS_SIGNATURE.byte[i])
                throw CFFileFormatException("Invalid OLE structured storage file");
        }
    }
};