/*
 "The contents of this file are subject to the Mozilla Public License
 
 Version 1.1 (the "License"); you may not use this file except in
 
 compliance with the License. You may obtain a copy of the License at
 
 http://www.mozilla.org/MPL/
 
 
 Software distributed under the License is distributed on an "AS IS"
 
 basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 
 License for the specific language governing rights and limitations
 
 under the License.
 
 
 The Original Code is OpenMCDF 1.5.4.
 
 
 The Initial Developer of the Original Code is Federico Blaseotto.
 
 Portions created by Federico Blaseotto are Copyright (C)2010 ~ 2013 Federico Blaseotto.
 
 All Rights Reserved.
 
 
 Contributor(s): Autodesk, Inc.."
*/


//
//  FItem.h
//  LIBCDF
//
//  Created by Weihua Luo on 4/3/2013.
//
#pragma once
#include "IFs/IFObject.h"
#include "IFs/IFComparable.h"
#include "FDirectoryEntry.h"

class CFCompoundFile;
class LIBCDF_EXPORT CFItem : public RefObject, public IFComparable
{
public:
	CFItem(void);
	virtual ~CFItem(void);

	virtual bool Equals(const IFObject* obj) const;
	virtual int CompareTo(const IFComparable* obj) const
	{
		const CFItem* pItem = dynamic_cast<const CFItem*>(obj);
		if (!pItem) return -1;
        return _dirEntry->CompareTo(pItem->DirEntry());
	}

	virtual void Delete();

public:
    //  expose these methods so that the IStorageImpl, IStreamImpl can reference
    void FileAddRef();
    void FileReleaseRef();
    
private:
	CFCompoundFile* _compoundFile;

protected:
	CFCompoundFile* CompoundFile() const
    {
        return _compoundFile;
    }

    void CheckDisposed() const;

    CFItem(CFCompoundFile* compoundFile);


private:
	IDirectoryEntry* _dirEntry;

public:
	PROPERTY_P(IDirectoryEntry*, DirEntry, _dirEntry);

    bool operator==(const CFItem& rightItem)
    {
        // Return true if the fields match:
        return CompareTo(&rightItem) == 0;
    }

    bool operator !=(const CFItem& rightItem)
    {
        return !(*this == rightItem);
    }

    bool Equals(const IFObject* obj)
    {
		const CFItem* pItem = dynamic_cast<const CFItem*>(obj);
        return CompareTo(pItem) == 0;
    }

    int GetHashCode()
    {
        return _dirEntry->GetHashCode();
    }

    /// <summary>
    /// Get entity name
    /// </summary>
	const wchar_t* Name() const
    {
        const wchar_t* n = _dirEntry->GetEntryName();
        if (n)
        {
            return n;
        }
        else
			return L"";
    }

    /// <summary>
    /// Size in bytes of the item. It has a valid value 
    /// only if entity is a stream, otherwise it is setted to zero.
    /// </summary>
    DLONG Size()
    {
		return _dirEntry->Size();
    }


    /// <summary>
    /// Return true if item is Storage
    /// </summary>
    /// <remarks>
    /// This check doesn't use reflection or runtime type information
    /// and doesn't suffer related performance penalties.
    /// </remarks>
    bool IsStorage()
    {
        return (_dirEntry->DEType() == eStgStorage);
    }

    /// <summary>
    /// Return true if item is a Stream
    /// </summary>
    /// <remarks>
    /// This check doesn't use reflection or runtime type information
    /// and doesn't suffer related performance penalties.
    /// </remarks>
    bool IsStream()
    {
		return _dirEntry->DEType() == eStgStream;
    }

    /// <summary>
    /// Return true if item is the Root Storage
    /// </summary>
    /// <remarks>
    /// This check doesn't use reflection or runtime type information
    /// and doesn't suffer related performance penalties.
    /// </remarks>
    bool IsRoot()
    {
        return _dirEntry->DEType() == eStgRoot;
    }

    /// <summary>
    /// Get/Set the Creation Date of the current item
    /// </summary>
    CFDateTime CreationDate()
    {
		CFDateTime dt;
		dt.FromBytes(_dirEntry->CreationDate().byte);
		return dt;
	}

	void CreationDate(const CFDateTime& time)
    {
		struct tm t;
		SysTime2Tm(time.GetSysTime(), t);
		time_t tt = mktime(&t);
		BYTE8 tvalue;
		memcpy(tvalue.byte, &tt, 8);
        if (_dirEntry->DEType() != eStgStream && _dirEntry->DEType() != eStgRoot)
			_dirEntry->CreationDate(tvalue);
        else
			throw CFException("Creation Date can only be set on storage entries");
    }

    CFDateTime ModifyDate()
    {
		CFDateTime dt;
		dt.FromBytes(_dirEntry->ModifyDate().byte);
		return dt;
	}

	void ModifyDate(const CFDateTime& time)
    {
		struct tm t;
		SysTime2Tm(time.GetSysTime(), t);
		time_t tt = mktime(&t);
		BYTE8 tvalue;
		memcpy(tvalue.byte, &tt, 8);
        if (_dirEntry->DEType() != eStgStream && _dirEntry->DEType() != eStgRoot)
			_dirEntry->ModifyDate(tvalue);
        else
			throw CFException("Creation Date can only be set on storage entries");
    }

    /// <summary>
    /// Get/Set Object class Guid for Root and Storage entries.
    /// </summary>
    void CLSID(const BYTE16& ID) const
    {
        if (_dirEntry->DEType() != eStgStream)
        {
            _dirEntry->StorageCLSID( ID );
        }
        else
			throw CFException("Object class GUID can only be set on Root and Storage entries");
    }

	BYTE16 CLSID()
	{
		return _dirEntry->StorageCLSID();
	}
 
};

typedef std::vector<CFItem*> CFItemVector;
