/*
 "The contents of this file are subject to the Mozilla Public License
 
 Version 1.1 (the "License"); you may not use this file except in
 
 compliance with the License. You may obtain a copy of the License at
 
 http://www.mozilla.org/MPL/
 
 
 Software distributed under the License is distributed on an "AS IS"
 
 basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 
 License for the specific language governing rights and limitations
 
 under the License.
 
 
 The Original Code is OpenMCDF 1.5.4.
 
 
 The Initial Developer of the Original Code is Federico Blaseotto.
 
 Portions created by Federico Blaseotto are Copyright (C)2010 ~ 2013 Federico Blaseotto.
 
 All Rights Reserved.
 
 
 Contributor(s): Autodesk, Inc.."
*/

//
//  FSector.h
//  LIBCDF
//
//  Created by Weihua Luo on 4/3/2013.
//
#pragma once
#include "IFs/IFObject.h"
#include "FileStream.h"
#include "MemoryStream.h"
#include <vector>
#include <queue>
#include <map>


class CFSector : public RefObject, public IFDisposable
{
public:
	CFSector(void);
	~CFSector(void);

    static int MINISECTOR_SIZE;// = 64;

    static int FREESECT;// ((int)0xFFFFFFFF)
    static int ENDOFCHAIN;// ((int)0xFFFFFFFE)
    static int FATSECT;// ((int)0xFFFFFFFD)
    static int DIFSECT;// ((int)0xFFFFFFFC)


	typedef enum _tagSectorType
	{
		Normal, Mini, FAT, DIFAT, RangeLockSector, Directory
	}ESectorType;

	virtual void Delete() 
	{
		delete this;
	}

	virtual bool Equals(const IFObject* obj) const;
public:

	IMP_PROPERTY(bool, DirtyFlag, _dirtyFlag);

    bool IsStreamed() const
    {
        if (_stream != NULL && _size != MINISECTOR_SIZE) return ((_id * _size) + _size < _stream->Length());
		return false;
    }

private:
	bool _dirtyFlag;// = false;
	int _size;// = 0;
    CFileStream* _stream;
	ESectorType _type;
	int _id;// = -1;
	CFByteArray _data;

	bool _loaded;
public:
	CFSector(int size, CFileStream* stream)
    {
        _size = size;
        _stream = stream;
		//_data.Size(_size);	//delay malloc
		_dirtyFlag = false;
		_type = CFSector::Normal;
		_id = -1;
		_loaded = false;
    }

    CFSector(int size, unsigned char data[])
    {
        _size = size;
		_data.Capacity(_size);
        _data = data;
        _stream = NULL;
		_dirtyFlag = false;
		_type = CFSector::Normal;
		_id = -1;
		_loaded = true;
    }

    CFSector(int size)
    {
        _size = size;
		//_data.Size(_size);
        _stream = NULL;
		_dirtyFlag = false;
		_type = CFSector::Normal;
		_id = -1;
		_loaded = false;
    }

	IMP_PROPERTY(ESectorType, Type, _type);
	IMP_PROPERTY(int, ID, _id);

	int Size() {return _size;}; 

    CFByteArray& GetData()
    {
		if (_data.Byte() && _loaded) return _data;

		if (_data.Byte() == NULL)
        {
			_data.Capacity(_size);
		}

        if (IsStreamed())
        {
            _stream->Seek((long)_size + (long)_id * (long)_size, CFileStream::eBegin);
			_stream->Read(_data.Byte(), 0, _size);
			_loaded = true;
        }

        return _data;
    }

    //public: void SetSectorData(byte[] b)
    //{
    //    data = b;
    //}

    //public: void FillData(byte b)
    //{
    //    if (data != NULL)
    //    {
    //        for (int i = 0; i < data.Length; i++)
    //        {
    //            data[i] = b;
    //        }
    //    }
    //}

    void ZeroData()
    {
		_data.ZeroData();
        _dirtyFlag = true;
    }

    void ReleaseData()
    {
		_data.Clear();
    }

private:
	virtual void Dispose()
	{
		ReleaseData();
        _dirtyFlag = false;
        _id = ENDOFCHAIN;
        _size = 0;
	}

};

class CFSectorPtr
{
public:
	CFSectorPtr() {_pointer = NULL;};
	CFSectorPtr(CFSector* p) : _pointer(NULL) {*this = p;};
	
	CFSectorPtr(const CFSectorPtr& sp) : _pointer(NULL) {*this = sp;};

	virtual ~CFSectorPtr() 
	{
		if (_pointer) _pointer->ReleaseRef();
		_pointer = NULL;
	}

	CFSectorPtr& operator=(const CFSectorPtr& pSrc)
	{
		if (_pointer) _pointer->ReleaseRef();
		if (pSrc.sector()) pSrc->AddRef();
		_pointer = pSrc.sector();
		return *this;
	}

	CFSectorPtr& operator=(CFSector* pSrc)
	{
		if (_pointer) _pointer->ReleaseRef();
		if (pSrc) pSrc->AddRef();
		_pointer = pSrc;
		return *this;
	}

	const CFSector* c_sector() const {return _pointer;}
	CFSector* sector() const {return _pointer;}
	CFSector* operator->() const{return _pointer;}
    CFSector& operator*() const{return *_pointer;}

private:
	CFSector* _pointer;
};

class BitConverter
{
public:

	static int ToInt32(const CFByteArray& src, int start);

};

typedef std::vector<CFSectorPtr> SectorChain;
typedef std::queue<CFSectorPtr> SectorQueue;

