/*
 "The contents of this file are subject to the Mozilla Public License
 
 Version 1.1 (the "License"); you may not use this file except in
 
 compliance with the License. You may obtain a copy of the License at
 
 http://www.mozilla.org/MPL/
 
 
 Software distributed under the License is distributed on an "AS IS"
 
 basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 
 License for the specific language governing rights and limitations
 
 under the License.
 
 
 The Original Code is OpenMCDF 1.5.4.
 
 
 The Initial Developer of the Original Code is Federico Blaseotto.
 
 Portions created by Federico Blaseotto are Copyright (C)2010 ~ 2013 Federico Blaseotto.
 
 All Rights Reserved.
 
 
 Contributor(s): Autodesk, Inc.."
*/

//
//  FStream.h
//  LIBCDF
//
//  Created by Weihua Luo on 4/3/2013.
//
#pragma once
#include "FStream.h"
#include "FItem.h"
#include "FSector.h"

#ifdef WIN32
#include <stdint.h>
#endif

class LIBCDF_EXPORT CFStream : public CFItem
{
private:
	CFByteArray _theWhole;
    
    unsigned int _cursor;
    
    bool    _dirtyFlag; //  Flag to indicate if the stream is dirty or not

	CFStream(void);
public:
	virtual ~CFStream(void);

    CFStream(CFCompoundFile* sectorManager);

    CFStream(CFCompoundFile* sectorManager, IDirectoryEntry* dirEntry)
        : CFItem(sectorManager)
        , _cursor(0)
    {
        if (dirEntry == NULL || dirEntry->SID() < 0)
			throw CFException("Attempting to add a CFStream using an unitialized directory");

        DirEntry(dirEntry);
    }
	void Clear();

    /// <summary>
    /// Set the data associated with the stream object.
    /// </summary>
    /// <example>
    /// <code>
    ///    byte[] b = new byte[]{0x0,0x1,0x2,0x3};
    ///    CompoundFile cf = new CompoundFile();
    ///    CFStream myStream = cf.RootStorage.AddStream("MyStream");
    ///    myStream.SetData(b);
    /// </code>
    /// </example>
    /// <param name="data">Data bytes to write to this stream</param>
    void SetData(CFByteArray& data);

    /// <summary>
    /// Append the provided data to stream data.
    /// </summary>
    /// <example>
    /// <code>
    ///    byte[] b = new byte[]{0x0,0x1,0x2,0x3};
    ///    byte[] b2 = new byte[]{0x4,0x5,0x6,0x7};
    ///    CompoundFile cf = new CompoundFile();
    ///    CFStream myStream = cf.RootStorage.AddStream("MyStream");
    ///    myStream.SetData(b); // here we could also have invoked .AppendData
    ///    myStream.AppendData(b2);
    ///    cf.Save("MyLargeStreamsFile.cfs);
    ///    cf.Close();
    /// </code>
    /// </example>
    /// <param name="data">Data bytes to append to this stream</param>
    /// <remarks>
    /// This method allows user to create stream with more than 2GB of data, 
    /// appending data to the end of existing ones.
    /// Large streams (>2GB) are only supported by CFS version 4.
    /// Append data can also be invoked on streams with no data in order
    /// to simplify its use inside loops.
    /// </remarks>
    void AppendData(CFByteArray& data);

    /// <summary>
    /// Get the data associated with the stream object.
    /// </summary>
    /// <example>
    /// <code>
    ///     CompoundFile cf2 = new CompoundFile("AFileName.cfs");
    ///     CFStream st = cf2.RootStorage.GetStream("MyStream");
    ///     byte[] buffer = st.GetData();
    /// </code>
    /// </example>
    /// <returns>Array of byte containing stream data</returns>
    /// <exception cref="T:OpenMcdf.CFDisposedException">
    /// Raised when the owner compound file has been closed.
    /// </exception>
    unsigned char* GetData();
	CFByteArray& GetByteArray();

    /// <summary>
    /// Get <paramref name="count"/> bytes associated with the stream object, starting from
    /// a provided <paramref name="offset"/>. When method returns, count will contain the
    /// effective count of bytes read.
    /// </summary>
    /// <example>
    /// <code>
    /// CompoundFile cf = new CompoundFile("AFileName.cfs");
    /// CFStream st = cf.RootStorage.GetStream("MyStream");
    /// int count = 8;
    /// // The stream is supposed to have a length greater than offset + count
    /// byte[] data = st.GetData(20, ref count);  
    /// cf.Close();
    /// </code>
    /// </example>
    /// <returns>Array of byte containing stream data</returns>
    /// <exception cref="T:OpenMcdf.CFDisposedException">
    /// Raised when the owner compound file has been closed.
    /// </exception>
    unsigned char* GetData(unsigned int offset, unsigned int& count);

    bool Read(void *pv, /*in, out*/ unsigned int& cb);
    bool Write(const void *pv, /*in, out*/ unsigned int& cb);
    uint64_t GetCursorPosition() const;
    
    void SetSize(unsigned int size);
    
    void Commit();
    
    /// Seek - moves the cursor
    /// dwOrigin:
    ///     SEEK_SET
    ///     SEEK_CUR
    ///     SEEK_END
    ///
    /// returns the new cursor position
    ///
    /// NOTE:
    ///     though the parameters are tagged as uint64_t, we actually only supports 32 bit here
    ///     we use uint64_t just for compatibility
    uint64_t Seek(unsigned int dwOrigin, int64_t offset);

    /// <summary>
    /// Copy data from an existing stream.
    /// </summary>
    /// <param name="input">A stream to read from</param>
    /// <remarks>
    /// Input stream is NOT closed after method invocation.
    /// </remarks>
    void CopyFrom(CFileStream* input);

    void CopyFrom(CFStream& stm);
    
    bool CopyFrom(CFStream& stm, /*in, out*/uint64_t& cbRead, /*out*/ uint64_t& written);
    
public:
    void SetModifiedFlag(bool bModified = true);
    bool IsDirty() const;
};

