/*
 "The contents of this file are subject to the Mozilla Public License
 
 Version 1.1 (the "License"); you may not use this file except in
 
 compliance with the License. You may obtain a copy of the License at
 
 http://www.mozilla.org/MPL/
 
 
 Software distributed under the License is distributed on an "AS IS"
 
 basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 
 License for the specific language governing rights and limitations
 
 under the License.
 
 
 The Original Code is OpenMCDF 1.5.4.
 
 
 The Initial Developer of the Original Code is Federico Blaseotto.
 
 Portions created by Federico Blaseotto are Copyright (C)2010 ~ 2013 Federico Blaseotto.
 
 All Rights Reserved.
 
 
 Contributor(s): Autodesk, Inc.."
*/


//
//  FStreamView.h
//  LIBCDF
//
//  Created by Weihua Luo on 4/3/2013.
//
#pragma once
#include "IFs/IFObject.h"
#include "FileStream.h"
#include "FSector.h"
#include <math.h>

class CFStreamView : public CFileStream
{
private:
	int _sectorSize;
    SectorChain* _sectorChain;
    CFileStream* _stream;

    long _position;

    UDLONG _length;
    BYTE4 _buf;


	CFStreamView(void);
public:
	~CFStreamView(void);

    CFStreamView(SectorChain* sectorChain, int sectorSize, CFileStream* stream)
    {
		_position = 0;
		SetSectorChain(sectorChain, sectorSize, stream);
    }

    CFStreamView(SectorChain* sectorChain, int sectorSize, DLONG length, SectorQueue* availableSectors, CFileStream* stream)
    {
		_position = 0;
		SetSectorChain(sectorChain, sectorSize, stream);
        AdjustLength(length, availableSectors);
    }


    CFStreamView(SectorChain* sectorChain, int sectorSize, DLONG length, CFileStream* stream)
    {
		_position = 0;
		SetSectorChain(sectorChain, sectorSize, stream);
        AdjustLength(length);
    }

    void SetSectorChain(SectorChain* sectorChain, int sectorSize, CFileStream* stream)
	{
		_position = 0;
        if (sectorChain == NULL)
            throw CFException("Sector Chain cannot be null");
        if (sectorSize <= 0)
			throw CFException("Sector size must be greater than zero");

        _sectorChain = sectorChain;
        _sectorSize = sectorSize;
        _stream = stream;
	}

    SectorChain* BaseSectorChain()
    {
        return _sectorChain;
    }

    bool CanRead()
    {
        return true;
    }

    bool CanSeek()
    {
        return true;
    }

    bool CanWrite()
    {
        return true;
    }

    void Flush()
    {

    }

    virtual UDLONG Length()
    {
		return _length;
    }

    long Position() const
    {
        return _position;
    }

	void Position(long value)
    {
        if (_position > _length - 1)
			throw CFArgumentOutOfRangeException("value");

        _position = value;
    }

    void Close()
    {
    }

    int ReadInt32()
    {
		Read(_buf.byte, 0, 4);
        return (int)(((_buf.byte[0] | (_buf.byte[1] << 8)) | (_buf.byte[2] << 16)) | (_buf.byte[3] << 24));
    }

    virtual int ReadByte()
	{
		Read(_buf.byte, 0, 1);
		return (int)_buf.byte[0];
	}

    int Read(unsigned char buffer[], int offset, int count)
    {
        int nRead = 0;
        int nToRead = 0;

		if (_sectorChain != NULL && _sectorChain->size() > 0)
        {
            // First sector
            int secIndex = (int)(_position / (long)_sectorSize);

            // Bytes to read count is the min between request count
            // and sector border

			nToRead = _MIN(
				(*_sectorChain)[0]->Size() - ((int)_position % _sectorSize),
                count);

            if (secIndex < _sectorChain->size())
            {
                CFBuffer::BlockCopy(
                    (*_sectorChain)[secIndex]->GetData(),
                    (int)(_position % _sectorSize),
                    buffer,
                    offset,
                    nToRead
                    );
            }

            nRead += nToRead;

            secIndex++;

            // Central sectors
            while (nRead < (count - _sectorSize))
            {
                nToRead = _sectorSize;

                CFBuffer::BlockCopy(
                    (*_sectorChain)[secIndex]->GetData(),
                    0,
                    buffer,
                    offset + nRead,
                    nToRead
                    );

                nRead += nToRead;
                secIndex++;
            }

            // Last sector
            nToRead = count - nRead;

            if (nToRead != 0)
            {
                CFBuffer::BlockCopy(
                    (*_sectorChain)[secIndex]->GetData(),
                    0,
                    buffer,
                    offset + nRead,
                    nToRead
                    );

                nRead += nToRead;
            }

            _position += nRead;

            return nRead;

        }
        else
            return 0;

    }

    int Seek(long offset, int origin)
    {
        switch (origin)
        {
		case CFileStream::eBegin:
                _position = offset;
                break;

		case CFileStream::eCurrent:
                _position += offset;
                break;

		case CFileStream::eEnd:
                _position = (long)(Length() - offset);
                break;
        }

        if (_position > _length)
        {
            AdjustLength(_position);
        }
        
        return _position;
    }

private:
	void AdjustLength(DLONG value)
    {
        AdjustLength(value, NULL);
    }

	void AdjustLength(DLONG value, SectorQueue* availableSectors);

public:
	void SetLength(DLONG value)
    {
        AdjustLength(value);
    }

	virtual void WriteByte(unsigned char value);

	virtual void Write(const unsigned char* buffer, int offset, int count);

	void Write(const CFByteArray& buffer, int offset, int count);
};

