/*
 "The contents of this file are subject to the Mozilla Public License
 
 Version 1.1 (the "License"); you may not use this file except in
 
 compliance with the License. You may obtain a copy of the License at
 
 http://www.mozilla.org/MPL/
 
 
 Software distributed under the License is distributed on an "AS IS"
 
 basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 
 License for the specific language governing rights and limitations
 
 under the License.
 
 
 The Original Code is OpenMCDF 1.5.4.
 
 
 The Initial Developer of the Original Code is Federico Blaseotto.
 
 Portions created by Federico Blaseotto are Copyright (C)2010 ~ 2013 Federico Blaseotto.
 
 All Rights Reserved.
 
 
 Contributor(s): Autodesk, Inc.."
 */

//
//  FCompoundFile.cpp
//  LIBCDF
//
//  Created by Weihua Luo on 4/3/2013.
//
#include "StdAfx.h"
#include "FCompoundFile.h"
#include "FItem.h"
#include "FStorage.h"
#include "FStream.h"
#include "FileStream.h"
#include "MemoryStream.h"

#define FLAT_WRITE      //  No optimization on the number of write operations

int CFItemComparer::Compare(CFItem* x, CFItem* y)
{
	if (!x || !y) return -1;
    // X CompareTo Y : X > Y --> 1 ; X < Y  --> -1
    return (x->DirEntry()->CompareTo(y->DirEntry()));

    //Compare X < Y --> -1
}

/// <summary>
/// Number of DIFAT entries in the _header
/// </summary>
int CFCompoundFile::HEADER_DIFAT_ENTRIES_COUNT = 109;

/// <summary>
/// Number of FAT entries in a DIFAT Sector
/// </summary>
int CFCompoundFile::DIFAT_SECTOR_FAT_ENTRIES_COUNT = 127;

/// <summary>
/// Sectors ID entries in a FAT Sector
/// </summary>
int CFCompoundFile::FAT_SECTOR_ENTRIES_COUNT = 128;

/// <summary>
/// Sector ID Size (int)
/// </summary>
int CFCompoundFile::SIZE_OF_SID = 4;

/// <summary>
/// Initial capacity of the flushing queue used
/// to optimize commit writing operations
/// </summary>
int CFCompoundFile::FLUSHING_QUEUE_SIZE = 6000;

/// <summary>
/// Maximum size of the flushing buffer used
/// to optimize commit writing operations
/// </summary>
int CFCompoundFile::FLUSHING_BUFFER_MAX_SIZE = 1024 * 1024 * 16;

/// <summary>
/// Create a blank, version 3 compound file.
/// Sector recycle is turned off to achieve the best reading/writing 
/// performance in most common scenarios.
/// </summary>
/// <example>
/// <code>
/// 
///     byte[] b = new byte[10000];
///     for (int i = 0; i &lt; 10000; i++)
///     {
///         b[i % 120] = (byte)i;
///     }
///
///     CompoundFile cf = new CompoundFile();
///     CFStream myStream = cf.RootStorage.AddStream("MyStream");
///
///     Assert.IsNotNull(myStream);
///     myStream.SetData(b);
///     cf.Save("MyCompoundFile.cfs");
///     cf.Close();
///     
/// </code>
/// </example>
CFCompoundFile::CFCompoundFile(void)
    : _dirty(false)
{
	_freeStream = false;
	_closeStream = true;
	_disposed = false;
    _sectorRecycle = false;
    _updateMode = eReadOnly;
    _eraseFreeSectors = false;

	_sourceStream = NULL;
	_validationExceptionEnabled = true;
	Init();

	_sectors = new CFSectorCollection();
	_directoryEntries = new CFDirectoryEntries();
    _header = new CFHeader();

    //_sectors.OnVer3SizeLimitReached += new Ver3SizeLimitReached(OnSizeLimitReached);
	_sectors->AttachCallback(this, &CFCompoundFile::OnSizeLimitReached);

    DIFAT_SECTOR_FAT_ENTRIES_COUNT = (GetSectorSize() / 4) - 1;
    FAT_SECTOR_ENTRIES_COUNT = (GetSectorSize() / 4);

    //Root -- 
    _rootStorage = new CFStorage(this);

    _rootStorage->DirEntry()->SetEntryName(L"Root Entry");
    _rootStorage->DirEntry()->DEType(eStgRoot);
    _rootStorage->DirEntry()->RBColor(eBlack);

    SetModifiedFlag(false);
}

CFCompoundFile::CFCompoundFile(const std::wstring& fileName)
    : _dirty(false)
{
    _freeStream = false;
    _closeStream = true;
    _disposed = false;
    
    _sectorRecycle = false;
    _updateMode = eReadOnly;
    _eraseFreeSectors = false;
    
    _sourceStream = NULL;
    
    _header = NULL;
    _rootStorage = NULL;
    _sectors = NULL;
    _directoryEntries = NULL;
    
    _validationExceptionEnabled = true;
    Init();
    
    LoadFile(fileName);
    SetModifiedFlag(false);
    
    DIFAT_SECTOR_FAT_ENTRIES_COUNT = (GetSectorSize() / 4) - 1;
    FAT_SECTOR_ENTRIES_COUNT = (GetSectorSize() / 4);
}

CFCompoundFile::CFCompoundFile(const std::wstring& fileName, ECFUpdateMode updateMode, bool sectorRecycle, bool eraseFreeSectors)
    : _dirty(false)
{
    _freeStream = false;
    _closeStream = true;
    _disposed = false;
    _sectorRecycle = sectorRecycle;
    _updateMode = updateMode;
    _eraseFreeSectors = eraseFreeSectors;
    
    _sourceStream = NULL;
    
    _header = NULL;
    _rootStorage = NULL;
    _sectors = NULL;
    _directoryEntries = NULL;
    
    _validationExceptionEnabled = true;
    Init();
    LoadFile(fileName);
    SetModifiedFlag(false);
    
    DIFAT_SECTOR_FAT_ENTRIES_COUNT = (GetSectorSize() / 4) - 1;
    FAT_SECTOR_ENTRIES_COUNT = (GetSectorSize() / 4);
}

CFCompoundFile::CFCompoundFile(ECFVersion cfsVersion, bool sectorRecycle, bool eraseFreeSectors)
    : _dirty(false)
{
	_freeStream = false;
	_closeStream = true;
	_disposed = false;
    _sectorRecycle = sectorRecycle;
	_eraseFreeSectors = eraseFreeSectors;
	_updateMode = eUpdate;

	_sourceStream = NULL;
	_validationExceptionEnabled = true;
	Init();

	_sectors = new CFSectorCollection();
	_directoryEntries = new CFDirectoryEntries();
    _header = new CFHeader((unsigned short)cfsVersion);

	_sectors->AttachCallback(this, &CFCompoundFile::OnSizeLimitReached);
            
    DIFAT_SECTOR_FAT_ENTRIES_COUNT = (GetSectorSize() / 4) - 1;
    FAT_SECTOR_ENTRIES_COUNT = (GetSectorSize() / 4);

    //Root -- 
    _rootStorage = new CFStorage(this);

    _rootStorage->DirEntry()->SetEntryName(L"Root Entry");
    _rootStorage->DirEntry()->DEType(eStgRoot);
    _rootStorage->DirEntry()->RBColor(eBlack);

    //InsertNewDirectoryEntry(rootStorage.DirEntry);

    SetModifiedFlag(false);
}

CFCompoundFile::CFCompoundFile(const std::wstring& fileName, ECFUpdateMode updateMode, bool sectorRecycle, bool eraseFreeSectors, bool noValidationException)
{
    _freeStream = false;
    _closeStream = true;
    _disposed = false;
    _validationExceptionEnabled = !noValidationException;
    
    _sectorRecycle = sectorRecycle;
    _updateMode = updateMode;
    _eraseFreeSectors = eraseFreeSectors;
    
    _sourceStream = NULL;
    
    _header = NULL;
    _rootStorage = NULL;
    _sectors = NULL;
    _directoryEntries = NULL;
    Init();
    LoadFile(fileName);
    
    DIFAT_SECTOR_FAT_ENTRIES_COUNT = (GetSectorSize() / 4) - 1;
    FAT_SECTOR_ENTRIES_COUNT = (GetSectorSize() / 4);

    SetModifiedFlag(false);
}

CFCompoundFile::CFCompoundFile(CFileStream* stream, ECFUpdateMode updateMode, bool sectorRecycle, bool eraseFreeSectors)
{
    _freeStream = false;
    _closeStream = true;
    _disposed = false;
    _sectorRecycle = sectorRecycle;
    _updateMode = updateMode;
    _eraseFreeSectors = eraseFreeSectors;
    _sourceStream = NULL;
    
    _header = NULL;
    _rootStorage = NULL;
    _sectors = NULL;
    _directoryEntries = NULL;
    
    _validationExceptionEnabled = true;
    Init();
    LoadStream(stream);
    
    DIFAT_SECTOR_FAT_ENTRIES_COUNT = (GetSectorSize() / 4) - 1;
    FAT_SECTOR_ENTRIES_COUNT = (GetSectorSize() / 4);

    SetModifiedFlag(false);
}

CFCompoundFile::CFCompoundFile(CFileStream* stream, ECFUpdateMode updateMode, bool sectorRecycle, bool eraseFreeSectors, bool noValidationException)
{
    _freeStream = false;
    _closeStream = true;
    _disposed = false;
    _validationExceptionEnabled = !noValidationException;
    
    _sectorRecycle = sectorRecycle;
    _updateMode = updateMode;
    _eraseFreeSectors = eraseFreeSectors;
    _sourceStream = NULL;
    
    _header = NULL;
    _rootStorage = NULL;
    _sectors = NULL;
    _directoryEntries = NULL;
    Init();
    LoadStream(stream);
    
    DIFAT_SECTOR_FAT_ENTRIES_COUNT = (GetSectorSize() / 4) - 1;
    FAT_SECTOR_ENTRIES_COUNT = (GetSectorSize() / 4);
    
    SetModifiedFlag(false);
}

CFCompoundFile::CFCompoundFile(CFileStream* stream)
{
    _freeStream = false;
    _closeStream = true;
    _disposed = false;
    
    _sectorRecycle = false;
    _updateMode = eReadOnly;
    _eraseFreeSectors = false;
    _sourceStream = NULL;
    
    _header = NULL;
    _rootStorage = NULL;
    _sectors = NULL;
    _directoryEntries = NULL;
    
    _validationExceptionEnabled = true;
    Init();
    LoadStream(stream);
    
    DIFAT_SECTOR_FAT_ENTRIES_COUNT = (GetSectorSize() / 4) - 1;
    FAT_SECTOR_ENTRIES_COUNT = (GetSectorSize() / 4);

    SetModifiedFlag(false);
}

CFCompoundFile::~CFCompoundFile(void)
{
	Dispose();
}

void CFCompoundFile::Init()
{
#ifndef FLAT_WRITE
    if (_updateMode == eUpdate)
        _buffer.Capacity(FLUSHING_BUFFER_MAX_SIZE);
#endif
    _lockObject = NULL;
    
    _transactionLockAdded = false;
    _lockSectorId = -1;
    _transactionLockAllocated = false;
}

void CFCompoundFile::Commit()
{
    if (_updateMode != eUpdate) return;
    if (!IsDirty()) return;
    
    Commit(false);
}

void CFCompoundFile::Commit(bool releaseMemory)
{
    if (_disposed)
        throw CFDisposedException("Compound File closed: cannot commit data");
    
    if (_updateMode != eUpdate)
        throw CFInvalidOperation("Cannot commit data in Read-Only update mode");
    
    //  If this is a new file, then just call the save method
    if ((_sourceStream == nullptr) &&
        !_fileName.empty())
    {
        Save(_fileName, false);
        return;
    }
    
    //try
    //{
#ifndef FLAT_WRITE
    
    int sId = -1;
    int sCount = 0;
    int bufOffset = 0;
#endif
    int sSize = GetSectorSize();
    
    if (_header->MajorVersion() != (unsigned short)eVer_3)
        CheckForLockSector();
    
    _sourceStream->Seek(0, CFileStream::eBegin);
    CFByteArray insBuffer(GetSectorSize());
    _sourceStream->Write(insBuffer.Byte(), 0, sSize);
    
    CommitDirectory();
    
    bool gap = true;
    
    
    for (int i = 0; i < _sectors->Count(); i++)
    {
#ifdef FLAT_WRITE
        
        //Note:
        //Here sectors should not be loaded dynamically because
        //if they are NULL it means that no change has involved them;
        
        CFSector* s = (*_sectors)[i].sector();
        
        if (s != NULL && s->DirtyFlag())
        {
            if (gap)
                _sourceStream->Seek((long)((long)(sSize) + (long)i * (long)sSize), CFileStream::eBegin);
            
            _sourceStream->Write(s->GetData().Byte(), 0, sSize);
            _sourceStream->Flush();
            s->DirtyFlag(false);
            gap = false;
            
        }
        else
        {
            gap = true;
            continue;
        }
        if (releaseMemory)
        {
            s->ReleaseData();
            s = NULL;
            (*_sectors)[i] = NULL;
            
            //GC.Collect();
        }
        
        
        
#else
        CFSector* s = (*_sectors)[i].sector();
        
        if (s != NULL && s->DirtyFlag() && _flushingQueue.size() < (int)(_buffer.Length() / sSize))
        {
            //First of a block of contiguous sectors, mark id, start enqueuing
            
            if (gap)
            {
                sId = s->ID();
                gap = false;
            }
            
            _flushingQueue.push(CFSectorPtr(s));
        }
        else
        {
            //Found a gap, stop enqueuing, flush a write operation
            
            gap = true;
            sCount = (int)_flushingQueue.size();
            
            if (sCount == 0) continue;
            
            bufOffset = 0;
            while (_flushingQueue.size() > 0)
            {
                CFSector* r = _flushingQueue.front().sector();
                CFBuffer::BlockCopy(r->GetData(), 0, _buffer, bufOffset, sSize);
                r->DirtyFlag(false);
                
                if (releaseMemory)
                {
                    r->ReleaseData();
                }
                
                bufOffset += sSize;
                _flushingQueue.pop();
            }
            
            _sourceStream->Seek(((long)sSize + (long)sId * (long)sSize), CFileStream::eBegin);
            _sourceStream->Write(_buffer.Byte(), 0, sCount * sSize);
            
            //Console.WriteLine("W - " + (int)(sCount * sSize ));
        }
#endif
    }
    
#ifndef FLAT_WRITE
    sCount = (int)_flushingQueue.size();
    bufOffset = 0;
    
    while (_flushingQueue.size() > 0)
    {
        CFSector* r = _flushingQueue.front().sector();
        CFBuffer::BlockCopy(r->GetData(), 0, _buffer, bufOffset, sSize);
        r->DirtyFlag(false);
        
        if (releaseMemory)
        {
            r->ReleaseData();
            r = NULL;
        }
        
        bufOffset += sSize;
        _flushingQueue.pop();
    }
    
    if (sCount != 0)
    {
        _sourceStream->Seek((long)sSize + (long)sId * (long)sSize, CFileStream::eBegin);
        _sourceStream->Write(_buffer.Byte(), 0, sCount * sSize);
        //Console.WriteLine("W - " + (int)(sCount * sSize));
    }
    
#endif
    
    // Seek to beginning position and save _header (first 512 or 4096 bytes)
    _sourceStream->Seek(0, CFileStream::eBegin);
    _header->Write(_sourceStream);
    
    _sourceStream->Flush();
    
    SetModifiedFlag(false);
    //}
    //catch (Exception ex)
    //{
    //    throw CFException("Internal error while committing data", ex);
    //}
}

void CFCompoundFile::Delete()
{
    Commit();
    delete this;
}

bool CFCompoundFile::Equals(const IFObject* obj) const
{
    return static_cast<const IFObject*>(this) == obj;
}

void CFCompoundFile::OnSizeLimitReached()
{
    
    CFSector* rangeLockSector = new CFSector(GetSectorSize(), _sourceStream);
    _sectors->Add(rangeLockSector);
    
    rangeLockSector->Type(CFSector::RangeLockSector);
    
    _transactionLockAdded = true;
    _lockSectorId = rangeLockSector->ID();
}

CFStorage* CFCompoundFile::RootStorage()
{
	CFStorage* p = dynamic_cast<CFStorage*>(_rootStorage);
    return p;
}

void CFCompoundFile::DoLoadChildren(BTSearchTree<CFItem>& bst, IDirectoryEntry* de)
{
    if (de->Child() != NOSTREAM )
    {
        if ((*_directoryEntries)[de->Child()]->DEType() == eStgInvalid) return;

        if ((*_directoryEntries)[de->Child()]->DEType() == eStgStream)
            bst.Add(new CFStream(this, (*_directoryEntries)[de->Child()]));
        else
            bst.Add(new CFStorage(this, (*_directoryEntries)[de->Child()]));

        LoadSiblings(bst, (*_directoryEntries)[de->Child()]);
    }
}


unsigned char* CFCompoundFile::GetData(CFStream* cFStream, long offset, int& count, CFByteArray& buffer)
{

    unsigned char* result = NULL;
    IDirectoryEntry* de = cFStream->DirEntry();

    count = (int)_MIN((long)(de->Size() - offset), (long)count);

    CFStreamView* sView = NULL;


    if (de->Size() < _header->MinSizeStandardStream())
    {
        sView = new CFStreamView(GetSectorChain(de->StartSector(), CFSector::Mini), CFSector::MINISECTOR_SIZE, de->Size(), _sourceStream);
    }
    else
    {

        sView = new CFStreamView(GetSectorChain(de->StartSector(), CFSector::Normal), GetSectorSize(), de->Size(), _sourceStream);
    }

	buffer.Capacity(count);
	result = buffer.Byte();
    //result = new byte[(int)(count)];

    sView->Seek(offset, CFileStream::eBegin);
    sView->Read(result, 0, buffer.Length());

	delete sView;	//CACHE IT?

	return result;
}

unsigned char* CFCompoundFile::GetData(CFStream* cFStream, CFByteArray& buffer)
{

    if (_disposed)
		throw CFDisposedException("Compound File closed: cannot access data");

    //byte[] result = NULL;

    IDirectoryEntry* de = cFStream->DirEntry();

    //IDirectoryEntry root = (*_directoryEntries)[0];

    if (de->Size() < _header->MinSizeStandardStream())
    {
		int nsize = (int)(de->Size());
		buffer.Capacity(nsize);

        CFStreamView* miniView
            = new CFStreamView(GetSectorChain(de->StartSector(), CFSector::Mini), CFSector::MINISECTOR_SIZE, nsize, _sourceStream);

		miniView->Read(buffer.Byte(),0,nsize);

        //BinaryReader br = new BinaryReader(miniView);
        //result = br.ReadBytes((int)de->Size());
        //br.Close();
		delete miniView;	//CACHE IT?
    }
    else
    {
        CFStreamView* sView
            = new CFStreamView(GetSectorChain(de->StartSector(), CFSector::Normal), GetSectorSize(), de->Size(), _sourceStream);

        //result = new byte[(int)de.Size];
		buffer.Capacity((int)de->Size());

		sView->Read(buffer.Byte(), 0, buffer.Length());
		delete sView;	//CACHE IT?
    }

	return buffer.Byte();
}

void CFCompoundFile::Load(CFileStream* stream)
{
	if (!stream || stream->IsReady() == false) 
	{
		throw CFException("Invalid file stream.");
	}

    try
    {
        _header = new CFHeader();
        _directoryEntries = new CFDirectoryEntries();

        _sourceStream = stream;

        _header->Read(stream);

        int n_sector = Ceiling(((double)(stream->Length() - GetSectorSize()) / (double)GetSectorSize()));

        if (stream->Length() > 0x7FFFFF0)
            _transactionLockAllocated = true;

        _sectors = new CFSectorCollection();
        //sectors = new ArrayList();
        for (int i = 0; i < n_sector; i++)
        {
            _sectors->Add(NULL);
        }
		_sectors->AttachCallback(this, &CFCompoundFile::OnSizeLimitReached);

        LoadDirectories();

        _rootStorage = new CFStorage(this, (*_directoryEntries)[0]);
    }
    catch (...)//(Exception)
    {
        if (stream != NULL)
            stream->Close();

        throw;
    }
}

void CFCompoundFile::LoadFile(const std::wstring& fileName)
{
    if (!_disposed)
    {
        Dispose();
#ifndef FLAT_WRITE
        if (_updateMode == eUpdate)
            _buffer.Capacity(FLUSHING_BUFFER_MAX_SIZE);
#endif
    }
    _disposed = false;
    
    _fileName = fileName;
    
    CFileStream* fs = NULL;
    
    try
    {
        if (_updateMode == eReadOnly)
        {
            fs = new CFileStream(fileName.c_str(), L"rb");//FileMode.Open, FileAccess.Read, FileShare.Read;
        }
        else
        {
            fs = new CFileStream(fileName.c_str(), L"rb+");//FileMode.Open, FileAccess.ReadWrite, FileShare.ReadWrite);
        }
        
        Load(fs);
        _freeStream = true;
    }
    catch (CFException& ex)
    {
        if (fs != NULL){
            fs->Close();
            delete fs;
            _sourceStream = NULL;
        }
        throw ex;
    }
}

void CFCompoundFile::LoadStream(CFileStream* stream)
{
    if (!_disposed)
    {
        Dispose();
#ifndef FLAT_WRITE
		if (_updateMode == eUpdate)
			_buffer.Capacity(FLUSHING_BUFFER_MAX_SIZE);
#endif
    }
    _disposed = false;
    
    if (stream == NULL)
        throw CFException("Stream parameter cannot be NULL");
    
    if (!stream->CanSeek())
        throw CFException("Cannot load a non-seekable Stream");
    
    
    stream->Seek(0, CFileStream::eBegin);
    
    Load(stream);
}

void CFCompoundFile::SetNormalSectorChain(SectorChain* sectorChain)
{
	if (!sectorChain) return ;

	SectorChain::iterator pos = sectorChain->begin();
    for (; pos != sectorChain->end(); pos++)
    {
		CFSector* s = (*pos).sector();
        if (s->ID() == -1)
        {
            _sectors->Add(s);
            s->ID(_sectors->Count() - 1);
        }
    }

    SetFATSectorChain(sectorChain);
}

void CFCompoundFile::FreeChain(SectorChain* sectorChain, bool zeroSector)
{
    CFByteArray ZEROED_SECTOR(GetSectorSize());
    
    SectorChain* pFAT
    = GetSectorChain(-1, CFSector::FAT);
    
    CFStreamView* pFATView
    = new CFStreamView(pFAT, GetSectorSize(), pFAT->size() * GetSectorSize(), _sourceStream);
    
    // Zeroes out sector data (if requested)
    
    if (zeroSector)
    {
        for (int i = 0; i < sectorChain->size(); i++)
        {
            CFSector* s = (*sectorChain)[i].sector();
            s->ZeroData();
        }
    }
    
    // Update FAT marking unallocated sectors
    for (int i = 0; i < sectorChain->size() - 1; i++)
    {
        int currentId = (*sectorChain)[i]->ID();
        int nextId = (*sectorChain)[i + 1]->ID();
        
        //AssureLength(FATView, Math.Max(currentId * SIZE_OF_SID, nextId * SIZE_OF_SID));
        
        pFATView->Seek(currentId * 4, CFileStream::eBegin);
        pFATView->Write(CFByteArray(&CFSector::FREESECT), 0, 4);
    }
    delete pFATView;	//CACHE IT?
}

void CFCompoundFile::SetMiniSectorChain(SectorChain* sectorChain)
{
	if (!sectorChain) return ;

    SectorChain* miniFAT
        = GetSectorChain(_header->FirstMiniFATSectorID(), CFSector::Normal);

    SectorChain* miniStream
        = GetSectorChain(RootEntry()->StartSector(), CFSector::Normal);

    CFStreamView* miniFATView
        = new CFStreamView(
            miniFAT,
            GetSectorSize(),
            _header->MiniFatSectorsNumber() * CFSector::MINISECTOR_SIZE,
            _sourceStream
            );

    CFStreamView* miniStreamView
        = new CFStreamView(
            miniStream,
            GetSectorSize(),
            _rootStorage->Size(),
            _sourceStream);

    // Set updated/new sectors within the ministream
    for (int i = 0; i < sectorChain->size(); i++)
    {
		CFSector* s = (*sectorChain)[i].sector();

        if (s->ID() != -1)
        {
            // Overwrite

            miniStreamView->Seek(CFSector::MINISECTOR_SIZE * s->ID(), CFileStream::eBegin);
            miniStreamView->Write(s->GetData(), 0, CFSector::MINISECTOR_SIZE);
        }
        else
        {
            // Allocate, position ministream at the end of already allocated
            // ministream's sectors

            miniStreamView->Seek((long)_rootStorage->Size(), CFileStream::eBegin);
            miniStreamView->Write(s->GetData(), 0, CFSector::MINISECTOR_SIZE);
            s->ID((int)(miniStreamView->Position() - CFSector::MINISECTOR_SIZE) / CFSector::MINISECTOR_SIZE);

            _rootStorage->DirEntry()->Size(miniStreamView->Length());
        }
    }

    // Update miniFAT
    for (int i = 0; i < sectorChain->size() - 1; i++)
    {
        int currentId = (*sectorChain)[i]->ID();
        int nextId = (*sectorChain)[i + 1]->ID();

        //AssureLength(miniFATView, Math.Max(currentId * SIZE_OF_SID, nextId * SIZE_OF_SID));

        miniFATView->Seek(currentId * 4, CFileStream::eBegin);
        miniFATView->Write(CFByteArray(&nextId), 0, 4);
    }

    //AssureLength(miniFATView, sectorChain[sectorChain->Count - 1].Id * SIZE_OF_SID);

    // Write End of Chain in MiniFAT
	miniFATView->Seek((*sectorChain)[sectorChain->size() - 1]->ID() * SIZE_OF_SID, CFileStream::eBegin);
    miniFATView->Write(CFByteArray(&CFSector::ENDOFCHAIN), 0, 4);

    // Update sector chains
    SetNormalSectorChain(miniStreamView->BaseSectorChain());
    SetNormalSectorChain(miniFATView->BaseSectorChain());

    //Update HEADER and root storage when ministream changes
	if (miniFAT->size() > 0)
    {
        _rootStorage->DirEntry()->StartSector((*miniStream)[0]->ID());
        _header->MiniFatSectorsNumber( (unsigned int)miniFAT->size());
        _header->FirstMiniFATSectorID( (*miniFAT)[0]->ID() );
    }

	delete miniFATView;//CACHE IT?
	delete miniStreamView;//CACHE IT?
}

void CFCompoundFile::FreeMiniChain(SectorChain* sectorChain, bool zeroSector)
{
	CFByteArray ZEROED_MINI_SECTOR(CFSector::MINISECTOR_SIZE);

    SectorChain* miniFAT
        = GetSectorChain(_header->FirstMiniFATSectorID(), CFSector::Normal);

    SectorChain* miniStream
        = GetSectorChain(RootEntry()->StartSector(), CFSector::Normal);

    CFStreamView* miniFATView
        = new CFStreamView(miniFAT, GetSectorSize(), _header->MiniFatSectorsNumber() * CFSector::MINISECTOR_SIZE, _sourceStream);

    CFStreamView* miniStreamView
        = new CFStreamView(miniStream, GetSectorSize(), _rootStorage->Size(), _sourceStream);

    // Set updated/new sectors within the ministream
    if (zeroSector)
    {
        for (int i = 0; i < sectorChain->size(); i++)
        {
			CFSector* s = (*sectorChain)[i].sector();

            if (s->ID() != -1)
            {
                // Overwrite
                miniStreamView->Seek(CFSector::MINISECTOR_SIZE * s->ID(), CFileStream::eBegin);
                miniStreamView->Write(ZEROED_MINI_SECTOR, 0, CFSector::MINISECTOR_SIZE);
            }
        }
    }

    // Update miniFAT
    for (int i = 0; i < sectorChain->size() - 1; i++)
    {
        int currentId = (*sectorChain)[i]->ID();
        int nextId = (*sectorChain)[i + 1]->ID();

        // AssureLength(miniFATView, Math.Max(currentId * SIZE_OF_SID, nextId * SIZE_OF_SID));

        miniFATView->Seek(currentId * 4, CFileStream::eBegin);
        miniFATView->Write(CFByteArray(CFSector::FREESECT), 0, 4);
    }

    //AssureLength(miniFATView, (*sectorChain)[sectorChain->Count - 1]->ID() * SIZE_OF_SID);

    // Write End of Chain in MiniFAT
    miniFATView->Seek((*sectorChain)[sectorChain->size() - 1]->ID() * SIZE_OF_SID, CFileStream::eBegin);
    miniFATView->Write(CFByteArray(&CFSector::FREESECT), 0, 4);

    // Update sector chains
    SetNormalSectorChain(miniStreamView->BaseSectorChain());
    SetNormalSectorChain(miniFATView->BaseSectorChain());

    //Update HEADER and root storage when ministream changes
	if (miniFAT->size())
    {
        _rootStorage->DirEntry()->StartSector( (*miniStream)[0]->ID());
        _header->MiniFatSectorsNumber((unsigned int)miniFAT->size());
        _header->FirstMiniFATSectorID((*miniFAT)[0]->ID());
    }
	delete miniFATView;//CACHE IT?
	delete miniStreamView;//CACHE IT?
}

SectorChain* CFCompoundFile::GetMiniSectorChain(int secID)
{
    SectorChain* result
            = new SectorChain();

    if (secID != CFSector::ENDOFCHAIN)
    {
        int nextSecID = secID;

        SectorChain* miniFAT = GetNormalSectorChain(_header->FirstMiniFATSectorID());
        SectorChain* miniStream = GetNormalSectorChain(RootEntry()->StartSector());

        CFStreamView* miniFATView
            = new CFStreamView(miniFAT, GetSectorSize(), _header->MiniFatSectorsNumber() * CFSector::MINISECTOR_SIZE, _sourceStream);

        CFStreamView* miniStreamView =
            new CFStreamView(miniStream, GetSectorSize(), _rootStorage->Size(), _sourceStream);

        //BinaryReader miniFATReader = new BinaryReader(miniFATView);

        nextSecID = secID;

        while (true)
        {
            if (nextSecID == CFSector::ENDOFCHAIN)
                break;

            CFSector* ms = new CFSector(CFSector::MINISECTOR_SIZE, _sourceStream);
            //byte[] temp = new byte[CFSector::MINISECTOR_SIZE];
			CFByteArray temp(CFSector::MINISECTOR_SIZE);

            ms->ID( nextSecID );
            ms->Type( CFSector::Mini );

            miniStreamView->Seek(nextSecID * CFSector::MINISECTOR_SIZE, CFileStream::eBegin);
			miniStreamView->Read(ms->GetData().Byte(), 0, CFSector::MINISECTOR_SIZE);

            result->push_back(ms);

            miniFATView->Seek(nextSecID * 4, CFileStream::eBegin);
			nextSecID = miniFATView->ReadInt32();//miniFATReader.ReadInt32();
        }
		delete miniFATView;//CACHE IT?
		delete miniStreamView;//CACHE IT?
    }
    return result;
}

void CFCompoundFile::DoLoadSiblings(BTSearchTree<CFItem>& bst, IDirectoryEntry* de)
{
    if (ValidateSibling(de->LeftSibling()))
    {
        // If there're more left siblings load them...
        DoLoadSiblings(bst, (*_directoryEntries)[de->LeftSibling()]);
    }

    if ((*_directoryEntries)[de->SID()]->DEType() == eStgStream)
        bst.Add(new CFStream(this, (*_directoryEntries)[de->SID()]));
    else if ((*_directoryEntries)[de->SID()]->DEType() == eStgStorage)
        bst.Add(new CFStorage(this, (*_directoryEntries)[de->SID()]));


    if (ValidateSibling(de->RightSibling()))
    {
        // If there're more right siblings load them...
        DoLoadSiblings(bst, (*_directoryEntries)[de->RightSibling()]);
    }
}

SectorQueue* CFCompoundFile::FindFreeSectors(CFSector::ESectorType sType)
{
    SectorQueue* freeList = new SectorQueue();

    if (sType == CFSector::Normal)
    {

        SectorChain* FatChain = GetSectorChain(-1, CFSector::FAT);
        CFStreamView* fatStream = new CFStreamView(FatChain, GetSectorSize(), _header->FatSectorsNumber() * GetSectorSize(), _sourceStream);

        int idx = 0;

        while (idx < _sectors->Count())
        {
            int id = fatStream->ReadInt32();

            if (id == CFSector::FREESECT)
            {
				if ((*_sectors)[idx].c_sector() == NULL)
                {
                    CFSector* s = new CFSector(GetSectorSize(), _sourceStream);
					s->ID(idx);
                    (*_sectors)[idx] = s;

                }

				freeList->push((*_sectors)[idx]);
            }

            idx++;
        }
		delete fatStream;	//CACHE IT?
    }
    else
    {
        SectorChain* miniFAT
            = GetSectorChain(_header->FirstMiniFATSectorID(), CFSector::Normal);

        CFStreamView* miniFATView
            = new CFStreamView(miniFAT, GetSectorSize(), _header->MiniFatSectorsNumber() * CFSector::MINISECTOR_SIZE, _sourceStream);

        SectorChain* miniStream
            = GetSectorChain(RootEntry()->StartSector(), CFSector::Normal);

        CFStreamView* miniStreamView
            = new CFStreamView(miniStream, GetSectorSize(), _rootStorage->Size(), _sourceStream);

        long ptr = 0;

        int nMinisectors = (int)(miniStreamView->Length() / CFSector::MINISECTOR_SIZE);

        while (ptr < nMinisectors)
        {
            //AssureLength(miniStreamView, (int)miniFATView.Length);

            int id = miniFATView->ReadInt32();
            ptr += 4;

            if (id == CFSector::FREESECT)
            {
                CFSector* ms = new CFSector(CFSector::MINISECTOR_SIZE, _sourceStream);
				CFByteArray temp(CFSector::MINISECTOR_SIZE);

                ms->ID((int)((ptr - 4) / 4));
                ms->Type(CFSector::Mini);

                miniStreamView->Seek(ms->ID() * CFSector::MINISECTOR_SIZE, CFileStream::eBegin);
				miniStreamView->Read(ms->GetData().Byte(), 0, CFSector::MINISECTOR_SIZE);

                //freeList.Enqueue(ms);
				freeList->push(ms);
            }
        }
		delete miniFATView;//CACHE IT?
		delete miniStreamView;//CACHE IT?
    }

    return freeList;
}

CFItemVector* CFCompoundFile::GetAllNamedEntries(std::wstring entryName)
{
    CFDirectoryEntries* r = FindDirectoryEntries(entryName);
    CFItemVector* result = new CFItemVector();

	CFDirectoryEntries::iterator pos;
    for (pos = r->begin(); pos != r->end(); pos++)
    {
		IDirectoryEntry* id = *pos;
        if (id->GetEntryName() == entryName && id->DEType() != eStgInvalid)
        {
            CFItem* i = (id->DEType() == eStgStorage) ? (CFItem*)new CFStorage(this, id) : (CFItem*)new CFStream(this, id);
			result->push_back(i);
        }
    }

    return result;
}

void CFCompoundFile::ReleaseAllEntries(CFItemVector* pEntries)
{
	if (!pEntries) return;
	CFItemVector::iterator pos = pEntries->begin();
	for (; pos != pEntries->end(); pos++){
		if ((*pos) != NULL) 
			(*pos)->ReleaseRef();
	}
	pEntries->clear();
}

void CFCompoundFile::ShrinkCompoundFile(CFileStream* s)
{
    CFCompoundFile* cf = new CFCompoundFile(s);

    if (cf->_header->MajorVersion() != (unsigned short)eVer_3)
        throw CFException("Current implementation of free space compression does not support version 4 of Compound File Format");

    //using (
	CFCompoundFile* tempCF = new CFCompoundFile((ECFVersion)cf->_header->MajorVersion(), cf->_sectorRecycle, cf->_eraseFreeSectors);//)
	if(cf->HasSourceStream())
    {
        tempCF->RootStorage()->CopyFrom(*cf->RootStorage());

        CMemoryStream* tmpMS = new CMemoryStream((int)cf->SourceStream().Length()); //This could be a problem for v4

        tempCF->Save(tmpMS);
      
		tempCF->Close();

        // If we were based on a writable stream, we update
        // the stream and do reload from the compressed one...

        s->Seek(0, CFileStream::eBegin);
        tmpMS->WriteTo(s);

        s->Seek(0, CFileStream::eBegin);
        s->SetLength(tmpMS->Length());

        tmpMS->Close();
        cf->Close(false);
    }
    
    delete tempCF;
    delete cf;
}

void CFCompoundFile::Dispose(bool disposing)
{
    try
    {
        if (!_disposed)
        {
            //lock (_lockObject)
            //{
                if (disposing)
                {
                    // Call from user code...

                    if (_sectors != NULL)
                    {
                        _sectors->Clear();
						delete _sectors;
                        _sectors = NULL;
                    }

					if(_rootStorage != NULL){
						_rootStorage->ReleaseRef();
						_rootStorage = NULL; // Some problem releasing resources...
					}

					if (_header != NULL)
					{
						delete _header;
						_header = NULL;
					}

					if (_directoryEntries != NULL)
					{
						CFDirectoryEntries::iterator pos = _directoryEntries->begin();
						for (; pos != _directoryEntries->end(); pos++) {
							if (*pos) delete (*pos);
						}
						_directoryEntries->clear();
						delete _directoryEntries;
						_directoryEntries = NULL;
					}

                    _fileName = L"";
                    _lockObject = NULL;	//Release the lock object.
#ifndef FLAT_WRITE
					_buffer.Clear();
#endif
                }

                if (_sourceStream != NULL) {
                    if (_closeStream) _sourceStream->Close();
					if (_freeStream) delete _sourceStream;
				}
				_sourceStream = NULL;
            //}
        }
    }
    catch(...)//finally
    {
    }
    _disposed = true;
}

void CFCompoundFile::CheckForLockSector()
{
    //If transaction lock has been added and not yet allocated in the FAT...
    if (_transactionLockAdded && !_transactionLockAllocated)
    {
        CFStreamView* fatStream = new CFStreamView(GetFatSectorChain(), GetSectorSize(), _sourceStream);
        
        fatStream->Seek(_lockSectorId * 4, CFileStream::eBegin);
        fatStream->Write(CFByteArray(&CFSector::ENDOFCHAIN), 0, 4);
        
        _transactionLockAllocated = true;
        delete fatStream; //CACHE IT?
    }
    
}
/// <summary>
/// Allocate space, setup sectors id and refresh _header
/// for the new or updated FAT sector chain.
/// </summary>
/// <param name="sectorChain">The new or updated generic sector chain</param>
void CFCompoundFile::SetFATSectorChain(SectorChain* sectorChain)
{
    SectorChain* fatSectors = GetSectorChain(-1, CFSector::FAT);
    CFStreamView* fatStream =
    new CFStreamView(
                     fatSectors,
                     GetSectorSize(),
                     _header->FatSectorsNumber() * GetSectorSize(), _sourceStream
                     );
    
    // Write FAT chain values --
    
    for (int i = 0; i < sectorChain->size() - 1; i++)
    {
        
        CFSector* sN = (*sectorChain)[i + 1].sector();
        CFSector* sC = (*sectorChain)[i].sector();
        
        int sID = sN->ID();
        fatStream->Seek(sC->ID() * 4, CFileStream::eBegin);
        fatStream->Write(CFByteArray(&sID), 0, 4);
    }
    
    fatStream->Seek((*sectorChain)[sectorChain->size() - 1]->ID() * 4, CFileStream::eBegin);
    fatStream->Write(CFByteArray(&CFSector::ENDOFCHAIN), 0, 4);
    
    // Merge chain to CFS
    SetDIFATSectorChain(fatStream->BaseSectorChain());
    delete fatStream; //CACHE IT?
}

/// <summary>
/// Setup the DIFAT sector chain
/// </summary>
/// <param name="FATsectorChain">A FAT sector chain</param>
void CFCompoundFile::SetDIFATSectorChain(SectorChain* sectorChain)
{
    // Get initial sector's count
    _header->FatSectorsNumber( (int)sectorChain->size());
    
    // Allocate Sectors
    CFSector* s = NULL;
    SectorChain::iterator pos = sectorChain->begin();
    
    for (; pos != sectorChain->end(); pos++)
    {
        s = (*pos).sector();
        if (s->ID() == -1)
        {
            _sectors->Add(s);
            s->ID(_sectors->Count() - 1);
            s->Type(CFSector::FAT);
        }
    }
    
    // Sector count...
    int nCurrentSectors = _sectors->Count();
    
    // Temp DIFAT count
    int nDIFATSectors = (int)_header->DifatSectorsNumber();
    
    if (sectorChain->size() > HEADER_DIFAT_ENTRIES_COUNT)
    {
        nDIFATSectors = Ceiling((double)(sectorChain->size() - HEADER_DIFAT_ENTRIES_COUNT) / DIFAT_SECTOR_FAT_ENTRIES_COUNT);
        nDIFATSectors = LowSaturation(nDIFATSectors - (int)_header->DifatSectorsNumber()); //required DIFAT
    }
    
    // ...sum with new required DIFAT sectors count
    nCurrentSectors += nDIFATSectors;
    
    // ReCheck FAT bias
    while (_header->FatSectorsNumber() * FAT_SECTOR_ENTRIES_COUNT < nCurrentSectors)
    {
        CFSector* extraFATSector = new CFSector(GetSectorSize(), _sourceStream);
        _sectors->Add(extraFATSector);
        
        extraFATSector->ID( _sectors->Count() - 1);
        extraFATSector->Type(CFSector::FAT);
        
        sectorChain->push_back(CFSectorPtr(extraFATSector));
        
        _header->IncreaseFatSectorsNumber();
        nCurrentSectors++;
        
        //... so, adding a FAT sector may induce DIFAT sectors to increase by one
        // and consequently this may induce ANOTHER FAT sector (TO-THINK: May this condition occure ?)
        if (nDIFATSectors * DIFAT_SECTOR_FAT_ENTRIES_COUNT <
            (_header->FatSectorsNumber() > HEADER_DIFAT_ENTRIES_COUNT ?
             _header->FatSectorsNumber() - HEADER_DIFAT_ENTRIES_COUNT :
             0))
        {
            nDIFATSectors++;
            nCurrentSectors++;
        }
    }
    
    
    SectorChain* difatSectors =
    GetSectorChain(-1, CFSector::DIFAT);
    
    CFStreamView* difatStream
    = new CFStreamView(difatSectors, GetSectorSize(), _sourceStream);
    
    // Write DIFAT Sectors (if required)
    // Save room for the following chaining
    for (int i = 0; i < sectorChain->size(); i++)
    {
        if (i < HEADER_DIFAT_ENTRIES_COUNT)
        {
            _header->DIFAT()[i] = (*sectorChain)[i]->ID();
        }
        else
        {
            // room for DIFAT chaining at the end of any DIFAT sector (4 bytes)
            if (i != HEADER_DIFAT_ENTRIES_COUNT && (i - HEADER_DIFAT_ENTRIES_COUNT) % DIFAT_SECTOR_FAT_ENTRIES_COUNT == 0)
            {
                CFByteArray temp(sizeof(int));
                temp.ZeroData();
                difatStream->Write(temp, 0, sizeof(int));
            }
            
            int nID = (*sectorChain)[i]->ID();
            difatStream->Write(CFByteArray(&nID), 0, sizeof(int));
            
        }
    }
    
    // Allocate room for DIFAT sectors
    for (int i = 0; i < difatStream->BaseSectorChain()->size(); i++)
    {
        CFSector* psec = (*(difatStream->BaseSectorChain()))[i].sector();
        if (psec->ID() == -1)
        {
            _sectors->Add(psec);
            psec->ID(_sectors->Count() - 1);
            psec->Type(CFSector::DIFAT);
        }
    }
    
    _header->DifatSectorsNumber( (unsigned int)nDIFATSectors);
    
    
    // Chain first sector
    if (difatStream->BaseSectorChain() != NULL && difatStream->BaseSectorChain()->size() > 0)
    {
        CFSector* psec = (*(difatStream->BaseSectorChain()))[0].sector();
        _header->FirstDIFATSectorID(psec->ID());
        
        // Update _header information
        _header->DifatSectorsNumber((unsigned int)difatStream->BaseSectorChain()->size());
        
        // Write chaining information at the end of DIFAT Sectors
        for (int i = 0; i < difatStream->BaseSectorChain()->size() - 1; i++)
        {
            int nID = (*(difatStream->BaseSectorChain()))[i + 1]->ID();
            CFBuffer::BlockCopy(
                                CFByteArray(&nID),
                                0,
                                (*(difatStream->BaseSectorChain()))[i]->GetData(),
                                GetSectorSize() - sizeof(int),
                                4);
        }
        
        CFBuffer::BlockCopy(
                            CFByteArray(&CFSector::ENDOFCHAIN),
                            0,
                            (*(difatStream->BaseSectorChain()))[difatStream->BaseSectorChain()->size() - 1]->GetData(),
                            GetSectorSize() - sizeof(int),
                            sizeof(int)
                            );
    }
    else
        _header->FirstDIFATSectorID(CFSector::ENDOFCHAIN);
    
    // Mark DIFAT Sectors in FAT
    CFStreamView* fatStream =
    new CFStreamView(sectorChain, GetSectorSize(), _header->FatSectorsNumber() * GetSectorSize(), _sourceStream);
    
    
    for (size_t i = 0; i < _header->DifatSectorsNumber(); i++)
    {
        fatStream->Seek((*(difatStream->BaseSectorChain()))[i]->ID() * 4, CFileStream::eBegin);
        fatStream->Write(CFByteArray(&CFSector::DIFSECT), 0, 4);
    }
    
    for (int i = 0; i < _header->FatSectorsNumber(); i++)
    {
        fatStream->Seek((*(fatStream->BaseSectorChain()))[i]->ID() * 4, CFileStream::eBegin);
        fatStream->Write(CFByteArray(&CFSector::FATSECT), 0, 4);
    }
    
    //fatStream->Seek(fatStream->BaseSectorChain[fatStream->BaseSectorChain()->size() - 1]->ID() * 4, CFileStream::eBegin);
    //fatStream->Write(BitConverter.GetBytes(CFSector::ENDOFCHAIN), 0, 4);
    
    _header->FatSectorsNumber( (int)fatStream->BaseSectorChain()->size());
    delete difatStream;	//CACHE IT?
    delete fatStream; //CACHE IT?
}


/// <summary>
/// Get the DIFAT Sector chain
/// </summary>
/// <returns>A list of DIFAT sectors</returns>
SectorChain* CFCompoundFile::GetDifatSectorChain()
{
    int validationCount = 0;
    
    SectorChain* result
    = new SectorChain();
    
    int nextSecID
    = CFSector::ENDOFCHAIN;
    
    if (_header->DifatSectorsNumber() != 0)
    {
        validationCount = (int)_header->DifatSectorsNumber();
        
        CFSector* s = (*_sectors)[_header->FirstDIFATSectorID()].sector();
        
        if (s == NULL) //Lazy loading
        {
            s = new CFSector(GetSectorSize(), _sourceStream);
            s->Type(CFSector::DIFAT);
            s->ID(_header->FirstDIFATSectorID());
            (*_sectors)[_header->FirstDIFATSectorID()] = s;
        }
        
        result->push_back(s);
        
        while (true && validationCount >= 0)
        {
            nextSecID = BitConverter::ToInt32(s->GetData(), GetSectorSize() - 4);
            
            // Strictly speaking, the following condition is not correct from
            // a specification point of view:
            // only ENDOFCHAIN should break DIFAT chain but
            // a lot of existing compound files use FREESECT as DIFAT chain termination
            if (nextSecID == CFSector::FREESECT || nextSecID == CFSector::ENDOFCHAIN) break;
            
            validationCount--;
            
            if (validationCount < 0)
            {
                Close();
                throw CFCorruptedFileException("DIFAT sectors count mismatched. Corrupted compound file");
            }
            
            s = (*_sectors)[nextSecID].sector();
            
            if (s == NULL)
            {
                s = new CFSector(GetSectorSize(), _sourceStream);
                s->ID(nextSecID);
                (*_sectors)[nextSecID] = s;
            }
            
            result->push_back(s);
        }
    }
    
    return result;
}

/// <summary>
/// Get the FAT sector chain
/// </summary>
/// <returns>List of FAT sectors</returns>
SectorChain* CFCompoundFile::GetFatSectorChain()
{
    int N_HEADER_FAT_ENTRY = 109; //Number of FAT sectors id in the _header
    
    SectorChain* result
    = new SectorChain();
    
    int nextSecID
    = CFSector::ENDOFCHAIN;
    
    SectorChain* difatSectors = GetDifatSectorChain();
    
    int idx = 0;
    
    // Read FAT entries from the _header Fat entry array (max 109 entries)
    while (idx < _header->FatSectorsNumber() && idx < N_HEADER_FAT_ENTRY)
    {
        nextSecID = _header->DIFAT()[idx];
        CFSector* s = (*_sectors)[nextSecID].sector();
        
        if (s == NULL)
        {
            s = new CFSector(GetSectorSize(), _sourceStream);
            s->ID(nextSecID);
            s->Type(CFSector::FAT);
            (*_sectors)[nextSecID] = s;
        }
        
        result->push_back(s);
        
        idx++;
    }
    
    //Is there any DIFAT sector containing other FAT entries ?
    if (difatSectors->size() > 0)
    {
        CFStreamView* difatStream
        = new CFStreamView
        (
         difatSectors,
         GetSectorSize(),
         _header->FatSectorsNumber() > N_HEADER_FAT_ENTRY ?
         (_header->FatSectorsNumber() - N_HEADER_FAT_ENTRY) * 4 :
         0,
         _sourceStream
         );
        
        CFByteArray nextDIFATSectorBuffer(4);
        
        difatStream->Read(nextDIFATSectorBuffer.Byte(), 0, 4);
        nextSecID = BitConverter::ToInt32(nextDIFATSectorBuffer, 0);
        
        int i = 0;
        int nFat = N_HEADER_FAT_ENTRY;
        
        while (nFat < _header->FatSectorsNumber())
        {
            if (difatStream->Position() == ((GetSectorSize() - 4) + i * GetSectorSize()))
            {
                difatStream->Seek(4, CFileStream::eCurrent);
                i++;
                continue;
            }
            
            CFSector* s = (*_sectors)[nextSecID].sector();
            
            if (s == NULL)
            {
                s = new CFSector(GetSectorSize(), _sourceStream);
                s->Type(CFSector::FAT);
                s->ID(nextSecID);
                (*_sectors)[nextSecID] = s;//UUU
            }
            
            result->push_back(s);
            
            difatStream->Read(nextDIFATSectorBuffer.Byte(), 0, 4);
            nextSecID = BitConverter::ToInt32(nextDIFATSectorBuffer, 0);
            nFat++;
        }
        delete difatStream;	//CACHE IT?
    }
    else
    {
        delete difatSectors;
    }
    
    return result;
    
}

/// <summary>
/// Get a standard sector chain
/// </summary>
/// <param name="secID">First SecID of the required chain</param>
/// <returns>A list of sectors</returns>
SectorChain* CFCompoundFile::GetNormalSectorChain(int secID)
{
    SectorChain* result = new SectorChain();
    
    int nextSecID = secID;
    
    SectorChain* fatSectors = GetFatSectorChain();
    
    CFStreamView* fatStream = new CFStreamView(fatSectors, GetSectorSize(), fatSectors->size() * GetSectorSize(), _sourceStream);
    
    while (true)
    {
        if (nextSecID == CFSector::ENDOFCHAIN) break;
        
        if (nextSecID >= _sectors->Count())
            throw CFCorruptedFileException("Next Sector ID reference an out of range sector.");//String.Format("Next Sector ID reference an out of range sector. NextID : {0} while sector count {1}", nextSecID, sectors.Count));
        
        CFSector* s = (*_sectors)[nextSecID].sector();
        if (s == NULL)
        {
            s = new CFSector(GetSectorSize(), _sourceStream);
            s->ID(nextSecID);
            s->Type(CFSector::Normal);
            (*_sectors)[nextSecID] = s;
        }
        
        result->push_back(s);
        
        fatStream->Seek(nextSecID * 4, CFileStream::eBegin);
        int next = fatStream->ReadInt32();
        
        if (next != nextSecID)
            nextSecID = next;
        else
            throw CFCorruptedFileException("Cyclic sector chain found. File is corrupted");
    }
    delete fatStream;	//CACHE IT?
    
    return result;
}

SectorChain* CFCompoundFile::GetSectorChain(int secID, CFSector::ESectorType chainType)
{
    
    switch (chainType)
    {
		case CFSector::DIFAT:
            return GetDifatSectorChain();
            
		case CFSector::FAT:
            return GetFatSectorChain();
            
		case CFSector::Normal:
            return GetNormalSectorChain(secID);
            
		case CFSector::Mini:
            return GetMiniSectorChain(secID);
            
        default:
            throw CFException("Unsupproted chain type");
    }
}

void CFCompoundFile::InsertNewDirectoryEntry(IDirectoryEntry* de)
{
    // If we are not adding an invalid dirEntry as
    // in a normal loading from file (invalid dirs MAY pad a sector)
    if (de != NULL)
    {
        // Find first available invalid slot (if any) to reuse it
        for (int i = 0; i < _directoryEntries->size(); i++)
        {
            if ((*_directoryEntries)[i]->DEType() == eStgInvalid)
            {
                (*_directoryEntries)[i] = de;//NEED TO RELEASE MEMORY...
                de->SID(i);
                return;
            }
        }
    }
    
    // No invalid directory entry found
    _directoryEntries->push_back(de);
    de->SID( (int)(_directoryEntries->size() - 1) );
}

/// <summary>
/// Reset a directory entry setting it to StgInvalid in the Directory.
/// </summary>
/// <param name="sid">Sid of the directory to invalidate</param>
void CFCompoundFile::ResetDirectoryEntry(int sid)
{
    CFDirectoryEntries::iterator pos = _directoryEntries->begin() + sid;
    if (pos != _directoryEntries->end())
    {
        if (*pos) delete *pos;
        *pos = new CFDirectoryEntry(eStgInvalid);
    }
}

bool CFCompoundFile::RefreshSIDs(BinaryTrees::BinaryTreeNode<CFItem>& node)
{
    if (node.Value() != NULL)
    {
        if (node.Left() != NULL && (node.Left()->Value()->DirEntry()->DEType() != eStgInvalid))
        {
            node.Value()->DirEntry()->LeftSibling( node.Left()->Value()->DirEntry()->SID() );
        }
        else
        {
            node.Value()->DirEntry()->LeftSibling(NOSTREAM);
        }
        
        if (node.Right() != NULL && (node.Right()->Value()->DirEntry()->DEType() != eStgInvalid))
        {
            node.Value()->DirEntry()->RightSibling(node.Right()->Value()->DirEntry()->SID());
        }
        else
        {
            node.Value()->DirEntry()->RightSibling(NOSTREAM);
        }
        return true;
    }
    return false;
}

BTSearchTree<CFItem>* CFCompoundFile::GetChildrenTree(int sid)
{
    ITComparer<CFItem>* pComp = new CFItemComparer();
    BTSearchTree<CFItem>* bst
    = new BTSearchTree<CFItem>(pComp);
    
    // Load children from their original tree.
    DoLoadChildren(*bst, (*_directoryEntries)[sid]);
    
    // Rebuild of (Red)-Black tree of entry children.
    bst->VisitTreeInOrder([&](BinaryTrees::BinaryTreeNode<CFItem>* node)->bool {
        if (node==NULL) return false;
        return this->RefreshSIDs(*node);});
    return bst;
}

void CFCompoundFile::LoadSiblings(BTSearchTree<CFItem>& bst, IDirectoryEntry* de)
{
    if (de->LeftSibling() != NOSTREAM)
    {
        // If there're more left siblings load them...
        DoLoadSiblings(bst, (*_directoryEntries)[de->LeftSibling()]);
    }
    
    if (de->RightSibling() != NOSTREAM)
    {
        // If there're more right siblings load them...
        DoLoadSiblings(bst, (*_directoryEntries)[de->RightSibling()]);
    }
}

bool CFCompoundFile::ValidateSibling(int sid)
{
    if (sid != NOSTREAM)
    {
        // if this siblings id does not overflow current list
        if (sid >= _directoryEntries->size())
        {
            if (_validationExceptionEnabled)
            {
                Close();
                throw CFCorruptedFileException("A Directory Entry references the non-existent sid number ");// + sid.ToString());
            }
            else
                return false;
        }
        
        //if this sibling is valid...
        if ((*_directoryEntries)[sid]->DEType() == eStgInvalid)
        {
            if (_validationExceptionEnabled)
            {
                Close();
                throw CFCorruptedFileException("A Directory Entry has a valid reference to an Invalid Storage Type directory");
            }
            else
                return false;
        }
        
        StgType eType = (*_directoryEntries)[sid]->DEType();
        if ((int)eType<0 || (int)eType>5)
        {
            if (_validationExceptionEnabled)
            {
                Close();
                throw CFCorruptedFileException("A Directory Entry has an invalid Storage Type");
            }
            else
                return false;
        }
        
        return true; //No fault condition encountered for sid being validated
    }
    
    return false;
}


/// <summary>
/// Load directory entries from compound file. Header and FAT MUST be already loaded.
/// </summary>
void CFCompoundFile::LoadDirectories()
{
    SectorChain* directoryChain
    = GetSectorChain(_header->FirstDirectorySectorID(), CFSector::Normal);
    
    if (_header->FirstDirectorySectorID() == CFSector::ENDOFCHAIN)
        _header->FirstDirectorySectorID((*directoryChain)[0]->ID());
    
    CFStreamView* dirReader
    = new CFStreamView(directoryChain, GetSectorSize(), directoryChain->size() * GetSectorSize(), _sourceStream);
    
    
    while (dirReader->Position() < directoryChain->size() * GetSectorSize())
    {
        CFDirectoryEntry* de
        = new CFDirectoryEntry(eStgInvalid);
        
        //We are not inserting dirs. Do not use 'InsertNewDirectoryEntry'
        de->Read(dirReader);
        _directoryEntries->push_back(de);
        de->SID( (int)_directoryEntries->size() - 1 );
    }
    delete dirReader;	//CACHE IT?
}

void CFCompoundFile::RefreshIterative(BinaryTrees::BinaryTreeNode<CFItem>* node)
{
    if (node == NULL) return;
    RefreshSIDs(*node);
    RefreshIterative(node->Left());
    RefreshIterative(node->Right());
}
void CFCompoundFile::CommitDirectory()
{
    const int DIRECTORY_SIZE = 128;
    
    SectorChain* directorySectors = GetSectorChain(_header->FirstDirectorySectorID(), CFSector::Normal);
    
    CFStreamView* sv = new CFStreamView(directorySectors, GetSectorSize(), 0, _sourceStream);
    
    CFDirectoryEntries::iterator pos = _directoryEntries->begin();
    for (; pos != _directoryEntries->end(); pos++)
    {
        (*pos)->Write(sv);
    }
    
    int delta = (int)_directoryEntries->size();
    
    while (delta % (GetSectorSize() / DIRECTORY_SIZE) != 0)
    {
        CFDirectoryEntry* dummy = new CFDirectoryEntry(eStgInvalid);
        dummy->Write(sv);
        delta++;
    }
    
    SectorChain::iterator spos = directorySectors->begin();
    for (; spos != directorySectors->end(); spos++)
    {
        (*spos)->Type(CFSector::Directory);
    }
    
    SetNormalSectorChain(directorySectors);
    
    _header->FirstDirectorySectorID((*directorySectors)[0]->ID());
    
    //Version 4 supports directory sectors count
    if (_header->MajorVersion() == 3)
    {
        _header->DirectorySectorsNumber(0);
    }
    else
    {
        _header->DirectorySectorsNumber((int)directorySectors->size());
    }
    delete sv; //CACHE IT?
}

bool CFCompoundFile::IsDirty() const
{   return _dirty;  }

void CFCompoundFile::SetModifiedFlag(bool bModified /* = true */)
{   _dirty = bModified; }

void CFCompoundFile::Save()
{
    if (!_fileName.empty())
    {
        Save(_fileName, false);
    }
}


/// <summary>
/// Saves the in-memory image of Compound File to a file.
/// </summary>
/// <param name="fileName">File name to write the compound file to</param>
/// <exception cref="T:OpenMcdf.std::exception">Raised if destination file is not seekable</exception>

void CFCompoundFile::Save(const std::wstring& fileName, bool bUpdateFileName)
{
    if (_disposed)
        throw CFException("Compound File closed: cannot save data");
    
    if (bUpdateFileName)
    {
        //  Update the file name, so that when we release the storage, we can update the file
        _fileName = fileName;
    }
    
    CFileStream* fs = NULL;
    
    try
    {
        fs = new CFileStream(fileName.c_str(), L"wb");
        Save(fs);
        fs->Flush();
        fs->Close();
        delete fs;
    }
    catch (std::exception&)
    {
        if (fs != NULL){ fs->Close(); delete fs; }
        throw CFException("Error saving file.");//("Error saving file [" + fileName + "]", ex);
    }
    catch(...)
    {
        if (fs != NULL){
            fs->Flush();
            fs->Close();
            delete fs;
        }
    }
    
    SetModifiedFlag(false);
}

/// <summary>
/// Saves the in-memory image of Compound File to a stream->
/// </summary>
/// <remarks>
/// Destination Stream must be seekable.
/// </remarks>
/// <param name="stream">The stream to save compound File to</param>
/// <exception cref="T:OpenMcdf.std::exception">Raised if destination stream is not seekable</exception>
/// <exception cref="T:OpenMcdf.std::exception">Raised if Compound File Storage has been already disposed</exception>
/// <example>
/// <code>
///    MemoryStream ms = new MemoryStream(size);
///
///    CompoundFile cf = new CompoundFile();
///    CFStorage st = cf.RootStorage.AddStorage("MyStorage");
///    CFStream sm = st.AddStream("MyStream");
///
///    byte[] b = new byte[]{0x00,0x01,0x02,0x03};
///
///    sm.SetData(b);
///    cf.Save(ms);
///    cf.Close();
/// </code>
/// </example>
void CFCompoundFile::Save(CFileStream* stream)
{
    if (_disposed)
        throw CFException("Compound File closed: cannot save data");
    
    if (!stream->CanSeek())
        throw CFException("Cannot save on a non-seekable stream");
    
    CheckForLockSector();
    int sSize = GetSectorSize();
    
    try
    {
        CFByteArray buf(sSize);
        stream->Write(buf.Byte(), 0, sSize);
        
        CommitDirectory();
        
        for (int i = 0; i < _sectors->Count(); i++)
        {
            CFSector* s = (*_sectors)[i].sector();
            
            if (s == NULL)
            {
                // Load source (unmodified) sectors
                // Here we have to ignore "Dirty flag" of
                // sectors because we are NOT modifying the source
                // in a differential way but ALL sectors need to be
                // persisted on the destination stream
                s = new CFSector(sSize, _sourceStream);
                s->ID(i);
                
                //(*_sectors)[i] = s;
            }
            
            
            stream->Write(s->GetData().Byte(), 0, sSize);
            
            //s.ReleaseData();
            
        }
        
        stream->Seek(0, CFileStream::eBegin);
        _header->Write(stream);
    }
    catch(...)//catch (Exception )
    {
        throw CFException("Internal error while saving compound file to stream ");
    }
}

void CFCompoundFile::SetData(CFItem* cfItem, CFByteArray& buffer)
{
    SetStreamData(cfItem, buffer);
}

/// <summary>
/// INTERNAL DEVELOPMENT. DO NOT CALL.
/// <param name="directoryEntry"></param>
/// <param name="buffer"></param>
void CFCompoundFile::AppendData(CFItem* cfItem, CFByteArray& buffer)
{
    //CheckFileLength();
    
    if (buffer.Byte() == NULL)
        printf("This special case maybe happen when reading old dataset!! \n");
    
    //Quick and dirty :-)
    if (buffer.Length() == 0) return;
    
    IDirectoryEntry* directoryEntry = cfItem->DirEntry();
    
    CFSector::ESectorType _st = CFSector::Normal;
    int _sectorSize = GetSectorSize();
    
    if (buffer.Length() + directoryEntry->Size() < _header->MinSizeStandardStream())
    {
        _st = CFSector::Mini;
        _sectorSize = CFSector::MINISECTOR_SIZE;
    }
    
    // Check for transition ministream -> stream:
    // Only in this case we need to free old sectors,
    // otherwise they will be overwritten.
    
    
    CFByteArray tempMini;
    
    if (directoryEntry->StartSector() != CFSector::ENDOFCHAIN)
    {
        if ((directoryEntry->Size() + buffer.Length()) > _header->MinSizeStandardStream() && directoryEntry->Size() < _header->MinSizeStandardStream())
        {
            tempMini.Capacity((int)directoryEntry->Size());
            
            CFStreamView* miniData
            = new CFStreamView(GetMiniSectorChain(directoryEntry->StartSector()), CFSector::MINISECTOR_SIZE, _sourceStream);
            
            miniData->Read(tempMini.Byte(), 0, (int)directoryEntry->Size());
            FreeMiniChain(GetMiniSectorChain(directoryEntry->StartSector()), _eraseFreeSectors);
            
            directoryEntry->StartSector( CFSector::ENDOFCHAIN );
            directoryEntry->Size(0);
            delete miniData;	//CACHE IT?
        }
    }
    
    SectorChain* sectorChain
    = GetSectorChain(directoryEntry->StartSector(), _st);
    
    SectorQueue* freeList = FindFreeSectors(_st); // Collect available free sectors
    
    CFStreamView* sv = new CFStreamView(sectorChain, _sectorSize, tempMini.Byte() != NULL ? buffer.Length() + tempMini.Length() : buffer.Length(), freeList, _sourceStream);
    
    // If stream was a ministream, copy ministream data
    // in the new stream
    if (tempMini.Byte() != NULL)
    {
        sv->Seek(0, CFileStream::eBegin);
        sv->Write(tempMini, 0, tempMini.Length());
    }
    else
    {
        sv->Seek((long)directoryEntry->Size(), CFileStream::eBegin);
    }
    
    // Write appended data
    sv->Write(buffer, 0, buffer.Length());
    
    switch (_st)
    {
		case CFSector::Normal:
            SetNormalSectorChain(sv->BaseSectorChain());
            break;
            
		case CFSector::Mini:
            SetMiniSectorChain(sv->BaseSectorChain());
            break;
        default:
            break;
    }
    
    
    if (sv->BaseSectorChain()->size() > 0)
    {
        directoryEntry->StartSector((*(sv->BaseSectorChain()))[0]->ID());
        directoryEntry->Size( (long)(buffer.Length()) + directoryEntry->Size());
        
        if (tempMini.Byte() != NULL)
        {
            int nSize = (int)directoryEntry->Size();
            nSize += tempMini.Length();
            directoryEntry->Size(nSize);
        }
    }
    else
    {
        directoryEntry->StartSector(CFSector::ENDOFCHAIN);
        directoryEntry->Size(0);
    }
    delete sv; //CACHE IT?
}

void CFCompoundFile::SetStreamData(CFItem* cfItem, CFByteArray& buffer)
{
    //CheckFileLength();
    if (buffer.Byte() == NULL)
        printf("This special case maybe happen when reading old dataset!! \n");
    
    //Quick and dirty :-)
    if (buffer.Length() == 0) return;
    
    IDirectoryEntry* directoryEntry = cfItem->DirEntry();
    
    CFSector::ESectorType _st = CFSector::Normal;
    int _sectorSize = GetSectorSize();
    
    if (buffer.Length() < _header->MinSizeStandardStream())
    {
        _st = CFSector::Mini;
        _sectorSize = CFSector::MINISECTOR_SIZE;
    }
    
    // Check for transition ministream -> stream:
    // Only in this case we need to free old sectors,
    // otherwise they will be overwritten.
    
    if (directoryEntry->StartSector() != CFSector::ENDOFCHAIN)
    {
        if (
            (buffer.Length() < _header->MinSizeStandardStream() && directoryEntry->Size() > _header->MinSizeStandardStream())
            || (buffer.Length() > _header->MinSizeStandardStream() && directoryEntry->Size() < _header->MinSizeStandardStream())
            )
        {
            
            if (directoryEntry->Size() < _header->MinSizeStandardStream())
            {
                FreeMiniChain(GetMiniSectorChain(directoryEntry->StartSector()), _eraseFreeSectors);
            }
            else
            {
                FreeChain(GetNormalSectorChain(directoryEntry->StartSector()), _eraseFreeSectors);
            }
            
            directoryEntry->Size(0);
            directoryEntry->StartSector(CFSector::ENDOFCHAIN);
        }
    }
    
    SectorChain* sectorChain = GetSectorChain(directoryEntry->StartSector(), _st);
    
    SectorQueue* freeList = NULL;
    
    if (_sectorRecycle)
        freeList = FindFreeSectors(_st); // Collect available free sectors
    
    CFStreamView* sv = new CFStreamView(sectorChain, _sectorSize, buffer.Length(), freeList, _sourceStream);
    
    sv->Write(buffer, 0, buffer.Length());
    
    switch (_st)
    {
		case CFSector::Normal:
            SetNormalSectorChain(sv->BaseSectorChain());
            break;
            
		case CFSector::Mini:
            SetMiniSectorChain(sv->BaseSectorChain());
            break;
        default:
            break;
    }
    
    
    if (sv->BaseSectorChain()->size() > 0)
    {
        directoryEntry->StartSector((*(sv->BaseSectorChain()))[0]->ID());
        directoryEntry->Size(buffer.Length());
    }
    else
    {
        directoryEntry->StartSector(CFSector::ENDOFCHAIN);
        directoryEntry->Size(0);
    }
    delete sv; //CACHE IT?
}

void CFCompoundFile::RemoveDirectoryEntry(int sid)
{
    if (sid >= _directoryEntries->size())
        throw CFException("Invalid SID of the directory entry to remove");
    
    if ((*_directoryEntries)[sid]->DEType() == eStgStream)
    {
        // Clear the associated stream (or ministream) if required
        if ((*_directoryEntries)[sid]->Size() > 0) //thanks to Mark Bosold for this !
        {
            if ((*_directoryEntries)[sid]->Size() < _header->MinSizeStandardStream())
            {
                SectorChain* miniChain
                = GetSectorChain((*_directoryEntries)[sid]->StartSector(), CFSector::Mini);
                FreeMiniChain(miniChain, _eraseFreeSectors);
            }
            else
            {
                SectorChain* chain
                = GetSectorChain((*_directoryEntries)[sid]->StartSector(), CFSector::Normal);
                FreeChain(chain, _eraseFreeSectors);
            }
        }
    }
    
    
    //Random r = new Random();
    CFRandom r;
    std::wstring sName = L"_DELETED_NAME_";
    sName += r.NextToWString(0x7FFF);
    (*_directoryEntries)[sid]->SetEntryName(sName);//"_DELETED_NAME_" + r.Next(short.MaxValue).ToString());
    (*_directoryEntries)[sid]->DEType(eStgInvalid);
}

CFDirectoryEntries* CFCompoundFile::FindDirectoryEntries(std::wstring entryName)
{
    CFDirectoryEntries* result = new CFDirectoryEntries();
    
    CFDirectoryEntries::iterator pos;
    for (pos = _directoryEntries->begin(); pos != _directoryEntries->end(); pos++)
    {
        IDirectoryEntry* p = *pos;
        if (p->GetEntryName() == entryName && p->DEType() != eStgInvalid)
            result->push_back(p);
    }
    
    return result;
}
