/*
 "The contents of this file are subject to the Mozilla Public License
 
 Version 1.1 (the "License"); you may not use this file except in
 
 compliance with the License. You may obtain a copy of the License at
 
 http://www.mozilla.org/MPL/
 
 
 Software distributed under the License is distributed on an "AS IS"
 
 basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 
 License for the specific language governing rights and limitations
 
 under the License.
 
 
 The Original Code is OpenMCDF 1.5.4.
 
 
 The Initial Developer of the Original Code is Federico Blaseotto.
 
 Portions created by Federico Blaseotto are Copyright (C)2010 ~ 2013 Federico Blaseotto.
 
 All Rights Reserved.
 
 
 Contributor(s): Autodesk, Inc.."
*/

//
//  FStorage.cpp
//  LIBCDF
//
//  Created by Weihua Luo on 4/3/2013.
//
#include "StdAfx.h"
#include "FStorage.h"
#include "FMock.h"
#include "FCompoundFile.h"

CFStorage::CFStorage(void)
{
	_children = NULL;
}

CFStorage::~CFStorage(void)
{
	if (_children) _children->Clear();
}

CFStorage::CFStorage(CFCompoundFile* compFile)
    : CFItem(compFile)
{
	_children = NULL;
    DirEntry( new CFDirectoryEntry(eStgStorage) );
	DirEntry()->RBColor( eBlack );
    compFile->InsertNewDirectoryEntry(DirEntry());
}

CFStorage::CFStorage(CFCompoundFile* compFile, IDirectoryEntry* dirEntry)
    : CFItem(compFile)
{
    if (dirEntry == NULL || dirEntry->SID() < 0)
		throw CFException("Attempting to create a CFStorage using an unitialized directory");
	_children = NULL;
    DirEntry( dirEntry );
}

BTSearchTree<CFItem>* CFStorage::LoadChildren(int SID)
{
    return CompoundFile()->GetChildrenTree(SID);
}

BTSearchTree<CFItem>* CFStorage::Children()
{
    // Lazy loading of _children tree.
    if (_children == NULL)
    {
        if (CompoundFile()->HasSourceStream())
        {
            _children = LoadChildren(DirEntry()->SID());
        }
        else
        {
            _children = new BTSearchTree<CFItem>();
        }
    }

    return _children;
}

bool CFStorage::Contains(const std::wstring &itemName)
{
	CheckDisposed();
    
	CFMock* tmp = new CFMock(itemName, eStgStream);

    bool f = Children()->Contains(tmp);
    tmp->ReleaseRef();

    return f;
}

CFStream* CFStorage::AddStream(const std::wstring& streamName)
{
    CheckDisposed();

	if (streamName.length() == 0)
        throw CFException("Stream name cannot be null or empty");

    CFStream* cfo = NULL;


    // Add new Stream directory entry
    cfo = new CFStream(CompoundFile());
    cfo->AddRef();
    cfo->DirEntry()->SetEntryName(streamName);


    try
    {
        // Add object to Siblings tree
        Children()->Add(cfo);

        //Rethread _children tree...
        CompoundFile()->RefreshIterative(Children()->Root());

        //... and set the root of the tree as new child of the current item directory entry
        DirEntry()->Child(Children()->Root()->Value()->DirEntry()->SID());
    }
	catch (BSTDuplicatedException&)
    {
        CompoundFile()->ResetDirectoryEntry(cfo->DirEntry()->SID());
        cfo->ReleaseRef();
        cfo = NULL;
		std::string mess = "An entry with name <";
		char* ps = WCS2MBS(streamName.c_str());
		if (ps){
			mess += ps; free(ps);
		}
		mess += "> is already present in storage <";
		ps = WCS2MBS(Name());
		if (ps){
			mess += ps; free(ps);
		}
		mess += ">";

        throw CFDuplicatedItemException(mess);//"An entry with name '" + streamName + "' is already present in storage '" + Name() + "' ");
    }

    CompoundFile()->SetModifiedFlag();
    
    return cfo;
}

CFStream* CFStorage::GetStream(const std::wstring& streamName)
{
	CheckDisposed();

	CFMock* tmp = new CFMock(streamName, eStgStream);

	//if (_children == null)
	//{
	//    _children = compoundFile.GetChildrenTree(SID);
	//}

	CFItem* outDe = NULL;

	if (Children()->TryFind(tmp, outDe) && outDe->DirEntry()->DEType() == eStgStream)
	{
		tmp->ReleaseRef();
		return (CFStream*)outDe;
	}
	else
	{
		tmp->ReleaseRef();
        return nullptr;
	}
}

CFStorage* CFStorage::GetStorage(const std::wstring& storageName)
{
    CheckDisposed();
    
    CFMock* tmp = new CFMock(storageName, eStgStorage);

    CFItem* outDe = NULL;
    if (Children()->TryFind(tmp, outDe) && outDe->DirEntry()->DEType() == eStgStorage)
    {
		tmp->ReleaseRef();
        return (CFStorage*)outDe;
    }
    else
    {
		tmp->ReleaseRef();
        return nullptr;
    }
}


CFStorage* CFStorage::AddStorage(const std::wstring& storageName)
{
    CheckDisposed();

	if (storageName.length()==0)
        throw CFException("Stream name cannot be null or empty");

    // Add new Storage directory entry
    CFStorage* cfo = NULL;

    cfo = new CFStorage(CompoundFile());
    cfo->AddRef();
    cfo->DirEntry()->SetEntryName(storageName);

    try
    {
        // Add object to Siblings tree
        Children()->Add(cfo);
    }
    catch (BSTDuplicatedException&)
    {

        CompoundFile()->ResetDirectoryEntry(cfo->DirEntry()->SID());
        cfo->ReleaseRef();
        cfo = NULL;
        throw CFDuplicatedItemException("An entry is already present in storage");//("An entry with name '" + storageName + "' is already present in storage '" + Name + "' ");
    }


    CompoundFile()->RefreshIterative(Children()->Root());
    DirEntry()->Child( Children()->Root()->Value()->DirEntry()->SID() );
    
    CompoundFile()->SetModifiedFlag();
    
    return cfo;
}

void CFStorage::Commit()
{   //  This is a simple implementation, that we just save the whole compound file
    if (CompoundFile() != nullptr)
    {
        CompoundFile()->Commit();
    }
}

void CFStorage::CopyFrom(CFStorage &src)
{
    //  CLSID
    CLSID(src.CLSID());
    
    src.VisitEntries([&](CFItem* item){
        if (item->IsStream())
        {
            CFStream* itemAsStream = (CFStream*)item;
            //  open the stream
            CFStream* st = GetStream(itemAsStream->Name());
            if (st == nullptr)
            {
                st = AddStream(itemAsStream->Name());
            }
            
            st->SetData(itemAsStream->GetByteArray());
        }
        else if (item->IsStorage())
        {
            CFStorage* itemAsStorage = (CFStorage*)item;
            CFStorage* strg = GetStorage(itemAsStorage->Name());
            if (strg == nullptr)
            {
                strg = (CFStorage*)AddStorage(itemAsStorage->Name());
            }
            strg->CopyFrom(*itemAsStorage);   // recursion, one level deeper
        }
	}, false);
    
    CompoundFile()->SetModifiedFlag();
}

bool CFStorage::DeleteItem(const std::wstring& entryName)
{
    CheckDisposed();

    // Find entry to delete
    CFMock* tmp = new CFMock(entryName, eStgInvalid);

    CFItem* foundObj = NULL;

    Children()->TryFind(tmp, foundObj);
	tmp->ReleaseRef();

    if (foundObj == NULL)
    {   return false;   }

    //if (foundObj.GetType() != typeCheck)
    //    throw CFException("Entry named [" + entryName + "] has not the correct type");

    if (foundObj->DirEntry()->DEType() == eStgRoot)
    {   return false;   }

	CFStorage* temp = NULL;
	ITEnumerator<CFItem>* penum = NULL;
    switch (foundObj->DirEntry()->DEType())
    {
        case eStgStorage:

            temp = (CFStorage*)foundObj;

			penum = temp->Children()->GetEnumeratorT();
			while(penum->MoveNext())
			{
				CFItem* de = penum->CurrentT();
				temp->DeleteItem(de->Name());
			}

            // Remove item from _children tree
            Children()->Remove(foundObj);

            // Synchronize tree with directory entries
            CompoundFile()->RefreshIterative(Children()->Root());

            // Rethread the root of siblings tree...
            if (Children()->Root() != NULL)
			{
                DirEntry()->Child(Children()->Root()->Value()->DirEntry()->SID());
			}
            else
			{
                DirEntry()->Child(NOSTREAM);
			}

            // ...and now remove directory (storage) entry
            CompoundFile()->RemoveDirectoryEntry(foundObj->DirEntry()->SID());

            break;

        case eStgStream:

            // Remove item from _children tree
            Children()->Remove(foundObj);

            // Synchronize tree with directory entries
            CompoundFile()->RefreshIterative(Children()->Root());

            // Rethread the root of siblings tree...
            if (Children()->Root() != NULL)
                DirEntry()->Child(Children()->Root()->Value()->DirEntry()->SID());
            else
                DirEntry()->Child(NOSTREAM);

            // Remove directory entry
            CompoundFile()->RemoveDirectoryEntry(foundObj->DirEntry()->SID());

            break;
        default:
            break;
    }

    CompoundFile()->SetModifiedFlag();

    //// Refresh recursively all SIDs (invariant for tree sorting)
    //VisitedEntryAction action = delegate(CFSItem target)
    //{
    //    if( ((IDirectoryEntry)target).SID>foundObj.SID )
    //    {
    //        ((IDirectoryEntry)target).SID--;
    //    }                   


    //    ((IDirectoryEntry)target).LeftSibling--;
    //};
    
    return true;
}

int CFStorage::RenameItem(const std::wstring &curName, const std::wstring &newName)
{
    CheckDisposed();
    
    // Find entry to rename
    CFMock* tmp = new CFMock(curName, eStgInvalid);
    
    CFItem* foundObj = NULL;
    
    Children()->TryFind(tmp, foundObj);
	tmp->ReleaseRef();
    
    if (foundObj == NULL)
    {   return 1;   }
    
    //  check if the element already exists
    tmp = new CFMock(newName, eStgInvalid);
    CFItem* shouldbenull = nullptr;
    Children()->TryFind(tmp, shouldbenull);
    if (shouldbenull != nullptr)
    {   return -1;  }       //  duplicate name
    
    //  Rename the element
    Children()->Remove(foundObj);
    foundObj->DirEntry()->SetEntryName(newName);
    Children()->Add(foundObj);
    
    return 0;
}