/*
 "The contents of this file are subject to the Mozilla Public License
 
 Version 1.1 (the "License"); you may not use this file except in
 
 compliance with the License. You may obtain a copy of the License at
 
 http://www.mozilla.org/MPL/
 
 
 Software distributed under the License is distributed on an "AS IS"
 
 basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 
 License for the specific language governing rights and limitations
 
 under the License.
 
 
 The Original Code is OpenMCDF 1.5.4.
 
 
 The Initial Developer of the Original Code is Federico Blaseotto.
 
 Portions created by Federico Blaseotto are Copyright (C)2010 ~ 2013 Federico Blaseotto.
 
 All Rights Reserved.
 
 
 Contributor(s): Autodesk, Inc.."
*/


//
//  FStream.cpp
//  LIBCDF
//
//  Created by Weihua Luo on 4/3/2013.
//
#include "StdAfx.h"
#include "FStream.h"
#include "FCompoundFile.h"

CFStream::CFStream(void)
    : _cursor(0)
    , _dirtyFlag(false)
{
}

CFStream::~CFStream(void)
{
	Clear();
}

void CFStream::Clear()
{
	_theWhole.Clear();
}

CFStream::CFStream(CFCompoundFile* sectorManager)
    : CFItem(sectorManager)
    , _cursor(0)
    , _dirtyFlag(false)
{
    DirEntry( new CFDirectoryEntry(eStgStream) );
    DirEntry()->RBColor(eBlack);

    sectorManager->InsertNewDirectoryEntry(DirEntry());
}

void CFStream::SetData(CFByteArray& data)
{
    CheckDisposed();

    CompoundFile()->SetData(this, data);
    
    //  Update the internal cache
    _theWhole = data;
    //  If the cursor exceeds the max, then adjust it accordingly
    if (_cursor > _theWhole.Length()) _cursor = _theWhole.Length();
}

void CFStream::AppendData(CFByteArray& data)
{
    CheckDisposed();
    if (Size() > 0)
    {
        CompoundFile()->AppendData(this, data);
    }
    else
    {
        CompoundFile()->SetData(this, data);
    }
}

unsigned char* CFStream::GetData()
{
    CheckDisposed();
		
    return CompoundFile()->GetData(this, _theWhole);
}

CFByteArray& CFStream::GetByteArray()
{
	if (_theWhole.Byte()) return _theWhole;
	GetData();
	return _theWhole;
}

unsigned char* CFStream::GetData(unsigned int offset, unsigned int& count)
{
    CheckDisposed();

    CFByteArray& arr = GetByteArray();
    
    //  return the data
    if (count > arr.Length() - offset)
    {
        count = arr.Length() - offset;
    }
    
    return arr.Byte() + offset;
}

void CFStream::CopyFrom(CFileStream* input)
{
    CheckDisposed();
	DLONG len = input->Length();
    if (input->CanSeek())
    {
        input->Seek(0, CFileStream::eBegin);
    }

	int ilen = 0;
	if (len > 0x7FFFFFFF) ilen = 0x7FFFFFFF;
	else ilen = (int)len;

	CFByteArray buffer(ilen);

	input->Read(buffer.Byte(), 0, ilen);
    SetData(buffer);	// the buffer will be released at the end of this method
	buffer.Clear();

	DLONG llen = ilen;
	while (llen < len)
	{
		if ((len - llen) > 0x7FFFFFFF) ilen = 0x7FFFFFFF;
		else ilen = (int)(len - llen);

		buffer.Capacity(ilen);
		input->Read(buffer.Byte(), 0, ilen);
		AppendData(buffer);
		buffer.Clear();

		llen += ilen;
	}
}

void CFStream::CopyFrom(CFStream &stm)
{
    _theWhole = stm.GetByteArray();
    //  If the cursor exceeds the max, then adjust it accordingly
    if (_cursor > _theWhole.Length()) _cursor = _theWhole.Length();
    
    //  Set it dirty
    SetModifiedFlag();

    CompoundFile()->SetModifiedFlag();
}

bool CFStream::CopyFrom(CFStream& stm, /*in, out*/uint64_t& cbRead, /*out*/ uint64_t& written)
{   //  so far, we only support the MAX to ULONG
    unsigned int cb = (unsigned int)cbRead;
    unsigned char* p = stm.GetData((unsigned int)stm.GetCursorPosition(), cb);
    if (p == nullptr) return false;
    cbRead = cb;
    //  move the cursor forward
    stm.Seek(SEEK_CUR, cb);
    if (!Write(p, cb)) return false;
    written = cb;
    
    //  Set it dirty
    SetModifiedFlag();
    
    CompoundFile()->SetModifiedFlag();

    return true;
}

bool CFStream::Read(void *pv, /*in, out*/ unsigned int& cb)
{
    unsigned char* p = GetData(_cursor, cb);
    if (p == nullptr)
    {
        cb = 0;
        return false;
    }
    
    //  Forward the cursor
    _cursor += cb;
    ::memcpy(pv, p, cb);
    return true;
}

bool CFStream::Write(const void *pv, /*in, out*/ unsigned int& cb)
{
#if 0   //  debug code
    if (::wcsstr(Name(), L"Htzye") != nullptr)
    {
        if (cb != 4)
        {
            printf("===== Write to: %s, length=%04x", WCS2MBS(Name()), cb);
        }
        else
        {
            printf("===== Write to: %s, length=%04x, *=%08X", WCS2MBS(Name()), cb, *(unsigned int*)pv);
        }
        getchar();
    }
#endif
    CFByteArray& arr = GetByteArray();
    arr.Write(_cursor, pv, cb);
    _cursor += cb;
    
    //  Set it dirty
    SetModifiedFlag();
    
    CompoundFile()->SetModifiedFlag();

    return true;        //  Looks like we should always succeed :-)
}

uint64_t CFStream::GetCursorPosition() const
{   return _cursor; }

void CFStream::SetSize(unsigned int size)
{
    //  Set the size, and adjust the cursor position accordingly
    GetByteArray().SetLength(size);
    if (_cursor > size) _cursor = size;
}

void CFStream::Commit()
{
    if (IsDirty() && (_theWhole.CByte() != nullptr))
    {
        CompoundFile()->SetData(this, _theWhole);
        
        // Clear the dirty flag
        SetModifiedFlag(false);
    }
}

uint64_t CFStream::Seek(unsigned int dwOrigin, int64_t offset)
{
    CFByteArray& arr = GetByteArray();
    
    switch (dwOrigin)
    {
        case SEEK_CUR:
            offset += _cursor;
            break;
            
        case SEEK_SET:
            break;
            
        case SEEK_END:
            offset += arr.Length();
            
        default:
            break;
    }
    
    //  Adjust _cursor
    if (offset < 0) _cursor = 0;
    else if (offset > arr.Length())
    {
        arr.SetLength(offset);
        _cursor = (unsigned int)offset;
    }
    else _cursor = (unsigned int)offset;
    
    return _cursor;
}

void CFStream::SetModifiedFlag(bool bModified/*=true*/)
{
    _dirtyFlag = bModified;
}

bool CFStream::IsDirty() const
{
    return _dirtyFlag;
}